/* -*- mode: c; c-basic-offset: 2; -*- */
/******** <lalVerbatim file="ExtTrigApplySearchCV"> ********
Author: Rahkola, R
$Id: ExtTrigApplySearch.c,v 1.5 2002/04/18 19:36:54 rrahkola Exp $
********* </lalVerbatim> ********/

/* Include I/O functions if we're in the debug stage */
#ifdef DEBUG_EXTTRIG
#include <stdio.h>
#endif

/* C functions */
#include <math.h>
#include <mpi.h>            /* Contains all MPI_* functions */

/* standard LAL functions*/
#include <lal/LALStdlib.h>
#include <lal/AVFactories.h>
#include <lal/Date.h>
#include <lal/StochasticCrossCorrelation.h>
#include <lal/Comm.h>       /* LAL MPI interface */

/* Associated headers */
#include <ExtTrig.h>
#include <ExtTrigApplySearch.h>

#ifdef __cplusplus
extern "C" {
#endif

NRCSID( EXTTRIGAPPLYSEARCHC, "$Id: ExtTrigApplySearch.c,v 1.5 2002/04/18 19:36:54 rrahkola Exp $");

/********************* local static function declarations ********************/

static void
ApplySearchMaster(
		  LALStatus            *const status,
		  LALSearchOutput      *const output,
		  const LALSearchInput *const input,
		  ETSearchParams       *const params,
		  MPI_Comm              const comm   );

static void
ApplySearchSlave(
		 LALStatus            *const status,
		 LALSearchOutput      *const output,
		 const LALSearchInput *const input,
		 ETSearchParams       *const params,
		 MPI_Comm                    comm   );

static void
ExtTrigCrossCorrelate(
		      LALStatus                     *const status,
		      REAL4WithUnits                *const ccStatistic,
		      ETDataSegment                 *const data,
		      RealFFTPlan                   *const fftPlan );

/************************* wrapperAPI function definitions  ******************/

/*
 * Function LALApplySearch
 *
 * Determines if this node is a master or slave and calls the
 * appropriate function.
 */
void
LALApplySearch(
    LALStatus             *status,
    LALSearchOutput       *output,
    LALSearchInput        *input,
    LALApplySearchParams  *applyParams)
{
  /* cast the ETSearch parameter structure */
  ETSearchParams* const params = (ETSearchParams *) applyParams->searchParams;

  INITSTATUS( status, "LALApplySearch", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "exttrig:LALApplySearch\n" );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

  /*
   * Check that output and parameters are configured correctly
   *
   */
  ASSERT( output, status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL );
  ASSERT( !(output->result), 
	  status, APPLYSEARCHH_ENNUL, APPLYSEARCHH_MSGENNUL );
  ASSERT( applyParams, status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL );
  ASSERT( applyParams->mpiParams,
	  status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL );
  ASSERT( applyParams->searchParams,
	  status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL );

  /*
   * Determine whether we're running as master or as a slave
   *
   */
  MPI_Comm_rank( *(applyParams->mpiParams->comm), &(params->rank) );

  if ( params->searchMaster ) /* master */
  {
    /* check that rank number confirms whether we're searchMaster */
    ASSERT( params->rank == 0,
	    status, APPLYSEARCHH_ERANK, APPLYSEARCHH_MSGERANK );
    ApplySearchMaster( status->statusPtr, output, input, params,
		       *(applyParams->mpiParams->comm)          );
  }
  else                        /* slave */
  {
    /* check that rank number confirms whether we're searchSlave */
    ASSERT( params->rank > 0,
	    status, APPLYSEARCHH_ERANK, APPLYSEARCHH_MSGERANK );
    ApplySearchSlave( status->statusPtr, output, input, params,
		      *(applyParams->mpiParams->comm)          );
  }
  CHECKSTATUSPTR( status );
  
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

/*
 * Function ApplySearchMaster:  
 * Called from LALApplySearch only on the search master node. It's main job is
 * to wait for responses for each slave ie. retrieve the statistics and wait
 * for the slaves to end, then make the output.  Since this module will run
 * continuously during the search, we wish to maximize the frequency for which
 * it checks if a search slave is finished and ready for more data. To this end
 * we need the following functionality:
 * 0a) If not initialized
 *    - initialize ApplySearchMaster
 * 1a) If all data is finished, and any slaves have finished
 *    - build the LALSearchOutput structure
 *    - FINISH
 * 1b) If not finished
 *    - keep output->notFinished set to TRUE
 * 2a) If no slaves
 *    - cross-correlate chunk 'n' of data (= next unprocessed chunk)
 *    - update SegsToBeAnalyzed
 * 2b) If slave(s) exist and are active
 *    - wait for slave message
 *  2b.a) If slave needs data
 *    - send SegsToBeAnalyzed
 *    - send data to slaves
 *    - update SegsToBeAnalyzed
 *    - FINISH
 *  2b.b) If slave has results
 *    - get results for chunk 'n'
 * 3) Plug results for chunk 'n' into ETResult
 * 4) Update SegsAnalyzed
 * 5) Update search status
 * 6) FINISH
 */
static void
ApplySearchMaster(
		  LALStatus            *const status,
		  LALSearchOutput      *const output,
		  const LALSearchInput *const input,
		  ETSearchParams       *const params,
		  MPI_Comm              const comm   )
{
  INT4                    i;           /* iterator(s)                       */
  MPIMessage              mpiMsg;      /* send/rec'v messages w/slave       */
  UINT4                   curSegment = /* index of first unanalyzed segment */
    params->data->filled - params->data->segsToBeAnalyzed;
  REAL4WithUnits          ccStatistic; /* calculated statistic              */
  REAL8                   tmpTime1,tmpTime2;  /* used for getting seg delay */
  INT2Vector             *tmpLALUnit = NULL;  /* temp storage for LALUnit   */
  /* pointers to various useful stuff */
  ETSegment              *segments = params->segments->data;
  ETResult               *results = params->result;
  ETDataSegment          *dataSeg = params->data->data;

  INITSTATUS( status, "ApplySearchMaster", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

#ifdef DEBUG_EXTTRIG
  fprintf( stdout, "exttrig:LALApplySearch - Master\n" );
#endif // DEBUG_EXTTRIG

  /*
   * 0a) If not initialized
   *    - initialize ApplySearchMaster
   */
  if (params->initialised == 0)
    {
      MPI_Comm_size(comm, &(params->numSlaves)); /* get the number of nodes */
      params->numSlaves -= 1;     /* subtract 1 to get the number of slaves */
      params->curSlaves  = params->numSlaves;  /* start out with all slaves */

      /* Allocate results vectors with only as many segments as we can find */
      results->xCorrValue = NULL;
      LALCreateVector( status->statusPtr, &(results->xCorrValue),
		       params->data->segsToBeAnalyzed );
      CHECKSTATUSPTR( status );
      for( i=0; i < results->xCorrValue->length; i++ ) 
	results->xCorrValue->data[i] = 0;

      results->xCorrDelay = NULL;
      LALCreateVector( status->statusPtr, &(results->xCorrDelay),
		       params->data->segsToBeAnalyzed );
      CHECKSTATUSPTR( status );
      for( i=0; i < results->xCorrDelay->length; i++ ) 
	results->xCorrDelay->data[i] = 0;

      results->xCorrLag = NULL;
      LALCreateVector( status->statusPtr, &(results->xCorrLag),
		       params->data->segsToBeAnalyzed );
      CHECKSTATUSPTR( status );
      for( i=0; i<results->xCorrLag->length;i++ ) results->xCorrLag->data[i]=0;

#ifdef DEBUG_EXTTRIG
      /* a prefix to identify debugging messages */
      snprintf( params->id, IDLEN, "master[node %2d]", params->rank );
      fprintf( stdout, "%s: Current number of slaves: %u\n",
	       params->id, params->numSlaves );
      fflush( stdout );
#endif // #ifdef DEBUG_EXTTRIG

      params->initialised = 1;          /* ApplySearchMaster is initialized */
    }
  
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, 
    "%s: Current segment = %d, segsToBeAnalyzed = %d, segsAnalyzed = %d\n",
	   params->id, curSegment, params->data->segsToBeAnalyzed,
	   params->data->segsAnalyzed );
  fprintf( stdout, "%s: Number of active slaves = %d\n", params->id,
	   params->curSlaves );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

  /*
   * 1a) If all data is finished, and any slaves have finished
   *    - build the LALSearchOutput structure
   *    - FINISH
   */
  if ( params->curSlaves == 0 && params->data->segsToBeAnalyzed == 0 )
    {
#ifndef DEBUG_EXTTRIG
      /* build output database entries */
      ExtTrigBuildOutput( status->statusPtr, output, params );
      CHECKSTATUSPTR( status );
#else
      output->numOutput     = 0;
      output->fracRemaining = 0.0;
      output->notFinished   = 0;
#endif // #ifndef DEBUG_EXTTRIG

#ifdef DUMMY_CODE
      fprintf( stdout, "%s: All done-- I'm finishing!\n", params->id );
      fflush( stdout );
#endif // DUMMY_CODE

      DETATCHSTATUSPTR( status );
      RETURN( status );
    }

  /*
   * 1b) If not finished
   *    - keep output->notFinished set to TRUE
   */
  else
    {
      output->notFinished   = 1;
      output->fracRemaining = (REAL4) (params->data->segsToBeAnalyzed / 
				       (REAL4) params->data->length    );
      output->numOutput     = 0;
    }

  /*
   * 2a) If no slaves
   *    - cross-correlate chunk 'n' of data (= next unprocessed chunk)
   *    - update SegsToBeAnalyzed
   */
  if ( !params->numSlaves )
    {
      ExtTrigCrossCorrelate( status->statusPtr, &ccStatistic, 
			     &(params->data->data[curSegment]), 
			     params->fftPlan                   );
      CHECKSTATUSPTR( status );

      params->data->segsToBeAnalyzed--;
    }

  /*
   * 2b) If slave(s) exist and are active
   *    - wait for slave message
   */
  else
    {
      LALMPIRecvMsg(status->statusPtr, &mpiMsg, comm);
      CHECKSTATUSPTR( status );

      /*
       *  2b.a) If slave needs data
       *    - send dataSegment ID to slaves
       *    - update SegsToBeAnalyzed
       *    - FINISH
       */
      if ( ! mpiMsg.send )
	{
	  /* update curSlaves */
	  if ( params->data->segsToBeAnalyzed == 0 )
	    {
	      curSegment++;
	      LALMPISendUINT4( status->statusPtr, &curSegment,
			       mpiMsg.source, comm            );
	      CHECKSTATUSPTR( status );

	      params->curSlaves--;
	    }
	  /* only send data if segsToBeAnalyzed > 0 */
	  else 
	    {
	      LALMPISendUINT4( status->statusPtr, &curSegment,
			       mpiMsg.source, comm            );
	      CHECKSTATUSPTR( status );

	      params->data->segsToBeAnalyzed--;
#if DEBUG_EXTTRIG > 1
	      fprintf( stdout, "%s: sent data to slave[node %2d]\n", 
		       params->id, mpiMsg.source );
	      fflush( stdout );
#endif // DEBUG_EXTTRIG > 1
	    }

	  DETATCHSTATUSPTR( status );
	  RETURN( status );
	}
      /*
       *  2b.b) If slave has results
       *    - get results for chunk 'n'
       */
      else if ( (mpiMsg.send == 1) )
	{
	  /* create temporary storage for LALUnit */
	  LALI2CreateVector( status->statusPtr, &tmpLALUnit, 
			     (1 + 2 * LALNumUnits)          );
	  CHECKSTATUSPTR( status );

	  LALMPIRecvUINT4( status->statusPtr, &curSegment, 
			   mpiMsg.source, comm            );
	  CHECKSTATUSPTR( status );

	  LALMPIRecvREAL4( status->statusPtr, &(ccStatistic.value),
			   mpiMsg.source, comm             );
	  CHECKSTATUSPTR( status );

	  LALMPIRecvINT2Vector( status->statusPtr, tmpLALUnit,
				mpiMsg.source, comm           );
	  CHECKSTATUSPTR( status );
	  
	  ccStatistic.units.powerOfTen = tmpLALUnit->data[0];
	  memcpy( ccStatistic.units.unitNumerator, 
		  &(tmpLALUnit->data[1]), 
		  LALNumUnits * sizeof(INT2)      );
	  memcpy( ccStatistic.units.unitDenominatorMinusOne, 
		  &(tmpLALUnit->data[1 + LALNumUnits]), 
		  LALNumUnits * sizeof(INT2)                );

	  /* destroy temporary storage for LALUnit */
	  LALI2DestroyVector( status->statusPtr, &tmpLALUnit );
	  CHECKSTATUSPTR( status );

#if DEBUG_EXTTRIG > 1
	      fprintf( stdout, "%s: received data from slave[node %2d]\n", 
		       params->id, mpiMsg.source );
	      fflush( stdout );
#endif // DEBUG_EXTTRIG > 1
	}
      else  /* Invalid MPI request from slave */
	{
	  ABORT( status, APPLYSEARCHH_ESEND, APPLYSEARCHH_MSGESEND );
	}
    }  // if ( params->curSlaves ) ...

  /*
   * 3) Plug results for chunk 'n' into ETResult
   */
  results->xCorrValue->data[curSegment] = ccStatistic.value;
  results->xCorrLag->data[curSegment] =
    segments[dataSeg[curSegment].segmentID].delay;
  LALGPStoFloat( status->statusPtr, &tmpTime1,
		 &(params->initParams->grbStart) );
  LALGPStoFloat( status->statusPtr, &tmpTime2,
		 &(segments[dataSeg[curSegment].segmentID].startGPS) );
  results->xCorrDelay->data[curSegment] = (REAL4)(tmpTime2 - tmpTime1);
  memcpy( &(results->xCorrUnit), &(ccStatistic.units), sizeof(LALUnit) );

#if DEBUG_EXTTRIG > 1
  fprintf( stdout, 
   "%s: dataSeg[%2d](=>seg%2d): value %8.3g,\tlag %8.6f,\tdelay %8.6f\n",
	   params->id,
	   curSegment, dataSeg[curSegment].segmentID, 
	   results->xCorrValue->data[curSegment],
	   results->xCorrLag->data[curSegment],
	   results->xCorrDelay->data[curSegment] );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

  /*
   * 4) Update SegsAnalyzed
   */
  params->data->segsAnalyzed++;
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: segsAnalyzed = %2d\n", params->id, 
	   params->data->segsAnalyzed                   );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

  /*
   * 5) Update search status
   */
  output->fracRemaining = 1.0 - 
    (REAL4) ( params->data->segsAnalyzed / (REAL4) params->data->length );

  /*
   * 6)  FINISH
   */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

/*
 * Function ApplySearchSlave
 *
 * Called by LALApplySearch only on slave nodes. Each slave node
 * calculates one stochastic cross-correlation at a time. Since there may be
 * no slave nodes in the final configuration, we should put most of the search
 * code into another function and use ApplySearchSlave as a wrapper for that
 * function.  To this end, we need the following functionality:
 * 0a) If not initialized
 *    - initialize ApplySearchSlave (debugging purposes)
 * 1)  Get data from master
 * 1a) If segsToBeAnalyzed == 0
 *    - update notFinished
 * 1b) If segsToBeAnalyzed > 0
 *    - set the current segment
 * 2)  Perform cross-correlation for chunk 'n' of data
 * 3)  Send results to master
 * 4)  FINISH
 */
static void
ApplySearchSlave(
		 LALStatus            *const status,
		 LALSearchOutput      *const output,
		 const LALSearchInput *const input,
		 ETSearchParams       *const params,
		 MPI_Comm                    comm   )
{
  UINT4                   curSegment;   /* index of segment being analyzed */
  MPIMessage              mpiMsg;       /* send/rec'v messages w/slave     */
  REAL4WithUnits          ccStatistic;  /* calculated statistic            */
  INT2Vector             *tmpLALUnit = NULL; /* temp storage for LALUnit   */

  INITSTATUS( status, "ApplySearchSlave", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

  /*
   * 0a) If not initialized
   *    - initialize ApplySearchSlave (debugging purposes)
   */
  if (params->initialised == 0)
    {
#ifdef DEBUG_EXTTRIG
      /* a prefix to identify debugging messages */
      snprintf( params->id, IDLEN, "slave [node %2d]", params->rank );
#endif // DEBUG_EXTTRIG

      params->initialised = 1;
    }

#ifdef DEBUG_EXTTRIG
      fprintf( stdout, "exttrig:LALApplySearch - %s\n", params->id );
      fflush( stdout );
#endif // DEBUG_EXTTRIG

  mpiMsg.msg    = 0;
  mpiMsg.source = params->rank;

  /*
   * 1)  Get data from master
   */
  mpiMsg.send   = 0; /* '0' = will receive data on next MPI exchange(s) */
  LALMPISendMsg( status->statusPtr, &mpiMsg, 0, comm );
  CHECKSTATUSPTR( status );

  LALMPIRecvUINT4( status->statusPtr, &curSegment, 0, comm );
  CHECKSTATUSPTR( status );

  /* check whether we need to perform any more analysis */
  if ( curSegment >= params->data->filled )
    {
      output->fracRemaining = 0.0;
      output->notFinished   = 0;
#ifdef DEBUG_EXTTRIG
      fprintf( stdout, "%s: finishing ApplySearch loop\n", params->id );
      fflush( stdout );
#endif // DUMMY_CODE
      DETATCHSTATUSPTR( status );
      RETURN( status );
    }
  
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: Current segment = %d\n", params->id, curSegment );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

  /*
   * 2)  Perform cross-correlation for chunk 'n' of data
   */
  ExtTrigCrossCorrelate( status->statusPtr, &ccStatistic, 
			 &(params->data->data[curSegment]), params->fftPlan );
  CHECKSTATUSPTR( status );

  /*
   * 3)  Send results to master
   */
  /* set up temporary storage for LALUnit information */
  LALI2CreateVector( status->statusPtr, &tmpLALUnit, 
		     (1 + 2 * LALNumUnits)          );
  CHECKSTATUSPTR( status );
  tmpLALUnit->data[0] =  ccStatistic.units.powerOfTen;
  memcpy( &(tmpLALUnit->data[1]), 
	  ccStatistic.units.unitNumerator, 
	  LALNumUnits * sizeof(INT2)      );
  memcpy( &(tmpLALUnit->data[1 + LALNumUnits]), 
	  ccStatistic.units.unitDenominatorMinusOne, 
	  LALNumUnits * sizeof(INT2)                );

  mpiMsg.send   = 1; /* '1' = will send data on next MPI exchange(s) */
  LALMPISendMsg( status->statusPtr, &mpiMsg, 0, comm );
  CHECKSTATUSPTR( status );

  LALMPISendUINT4( status->statusPtr, &curSegment, 0, comm );
  CHECKSTATUSPTR( status );
  
  LALMPISendREAL4( status->statusPtr, &(ccStatistic.value), 0, comm );
  CHECKSTATUSPTR( status );

  LALMPISendINT2Vector( status->statusPtr, tmpLALUnit, 0, comm );
  CHECKSTATUSPTR( status );

  /* destroy temporary storage for LALUnit */
  LALI2DestroyVector( status->statusPtr, &tmpLALUnit );
  CHECKSTATUSPTR( status );

  /*
   * 4)  FINISH
   */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

/*
 * Function ExtTrigCrossCorrelate:
 *
 * Called by either ApplySearchMaster or ApplySearchSlave.  This executes the
 * main search sequence-- that is, the StochasticCrossCorrelation() function.
 * It returns cross-correlation statistic via the REAL4WithUnits structure.  
 * To this end, we need the following functionality:
 * 1)  Allocate temporary memory for frequency series data
 * 2)  Zero-pad and FFT the time series data
 * 3)  Calculate the cross-correlation statistic
 * 4)  Clean up and FINISH
 */
static void
ExtTrigCrossCorrelate(
		      LALStatus                     *const status,
		      REAL4WithUnits                *const ccStatistic,
		      ETDataSegment                 *const data,
		      RealFFTPlan                   *const fftPlan )
{
  COMPLEX8FrequencySeries          fsChan1;   /* for LALStochasticC..C...    */
  COMPLEX8FrequencySeries          fsChan2;   /* for LALStochasticC..C...    */
  StochasticCrossCorrelationInput  ccInput;   /* for LALStochasticC..C...    */
    

  INITSTATUS( status, "ExtTrigCrossCorrelate", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

#ifdef DEBUG_EXTTRIG
  fprintf( stdout, "exttrig:ExtTrigCrossCorrelate\n" );
  fflush( stdout );
#endif // DEBUG_EXTTRIG

  /*
   * 1)  Allocate temporary memory for frequency series data
   */
  fsChan1.data = NULL;
  LALCCreateVector( status->statusPtr, &(fsChan1.data), 
		    data->filter->data->length         );
  CHECKSTATUSPTR( status );

  fsChan2.data = NULL;
  LALCCreateVector( status->statusPtr, &(fsChan2.data), 
		    data->filter->data->length         );
  CHECKSTATUSPTR( status );

  /*
   * 2)  Zero-pad and FFT the time series data
   */
  LALSZeroPadAndFFT( status->statusPtr, &fsChan1, data->tsChan1, fftPlan );
  CHECKSTATUSPTR( status );

  LALSZeroPadAndFFT( status->statusPtr, &fsChan2, data->tsChan2, fftPlan );
  CHECKSTATUSPTR( status );

  /* :KLUGE: set the filter->deltaF = fsChan1.deltaF for fixing CoarseGrain */
  data->filter->deltaF = fsChan1.deltaF;

#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "ExtTrigCrossCorrelate stats: \n" );
  fprintf( stdout, "   TSeries length = %d;\t deltaT = %g;\t f0 = %g\n",
	   data->tsChan1->data->length, data->tsChan1->deltaT, 
	   data->tsChan1->f0 );
  fprintf( stdout, "   FSeries length = %d;\t deltaF = %g;\t f0 = %g\n",
	   fsChan1.data->length, fsChan1.deltaF, fsChan1.f0 );
  fprintf( stdout, "   Filter  length = %d;\t deltaF = %g;\t f0 = %g\n",
	   data->filter->data->length, data->filter->deltaF, 
	   data->filter->f0 );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

  /*
   * 3)  Calculate the cross-correlation statistic
   */
  ccInput.hBarTildeOne = &fsChan1;
  ccInput.hBarTildeTwo = &fsChan2;
  ccInput.optimalFilter = data->filter;

  LALStochasticCrossCorrelationStatistic( status->statusPtr, ccStatistic,
					  &ccInput, 0                    );
  //  ccStatistic->value = 5.0;
  //  ccStatistic->units = data->tsChan1->sampleUnits;

  CHECKSTATUSPTR( status );

  /*
   * 4)  Clean up and FINISH
   */
  LALCDestroyVector( status->statusPtr, &(fsChan2.data) );
  CHECKSTATUSPTR( status );
		     
  LALCDestroyVector( status->statusPtr, &(fsChan1.data) );
  CHECKSTATUSPTR( status );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

#ifdef __cplusplus
}
#endif
