/******** <lalVerbatim file="KnownPulsarDemodCV"> ********
Author: Mendell, Greg A.
$Id: KnownPulsarDemod.c,v 1.11 2002/05/14 17:35:39 gmendell Exp $
********* </lalVerbatim> ********/

/* Notes: */
/*  DFT  = Discrete Fourier Transform, e.g., as computed by FFT routines. */
/*  SFT  = short time baseline DFT;  No demodulation of the data is done */
/*  DeFT = stitched together SFTs forming demodulated longer time baseline DFT. */
/*  DeFTs remove phase modulation due to Earth's motion and that intristic to the source */
/*  Eventually DeFTs will account for amplitude modulation due to beam pattern sensitivity of the detector */

/* 09/26/01 Code first submitted to lalwrapper cvs. */
/* Revisons: */
/*  09/27/01 gam Debugged code to produce correct SFT output.  Was adding data to multiDimData using complex_8; needed complex_16.; */
/*  09/28/01 gam Produce SFTs using LALForwardRealFFT rather than LALCOMPLEX8VectorFFT. */
/*  10/08/01 gam Copy in latest tdb function form LALDemodTest; update definitions and calls to funcName. */
/*  10/10/01 gam Change deltaT and tSFT to type REAL8; Check that input data interval and sample rate agrees with that requested; get units from input data. */
/*  10/10/01 gam Add error checking that data is found (unless generating test data) */
/*  10/10/01 gam Set up time intervals using input params->gpsStartTime */
/*  10/11/01 gam Add nSFT and mCohSFTPerCall as command line filter parameters.  */
/*  10/11/01 gam Clear up confusion between using N - 1 and N to compute tSFT, tCoh, duration, etc.... */
/*  NOTE THE FOLLOWING ASSUMPTIONS MUST BE TRUE FOR THIS CODE TO WORK */
/*  tSFT = Time between start of one SFT and next SFT for sequential SFTs */
/*  The difference between the times associated with the first and last data points in an n second frame file must correspond to n - deltaT seconds */
/*  The number of data points in the file is N = (n - deltaT)/deltaT + 1 = n/deltaT = n*sampleRate */
/*  The difference between the gpsStartTime and gpsStopTime in a frame file is deltaT*N not deltaT*(N - 1) */
/*  For DFTs that df = 1/N*deltaT */
/*  10/12/01 gam Move declaration and allocation of memory into InitSearch and ConditionData; Deallocate when finished with variable or in FinalizeSearch */
/*  10/15/01 gam Move option to generate test data into ConditionData. */
/*  10/15/01 gam Move declaration of variables used only in LALDemod into LALDemod section */
/*  10/15/01 gam Add code to compute mCohSFTPerCall per call to ApplySearch; return correct output->fracRemaining each time */
/*  10/16/01 gam Do not set up creator_db, process_id, or event_id; input filter_id, ifo, and target_name as command line filter parameters */
/*  10/17/01 gam Make duration database column type REAL4. */
/*  10/18/01 gam Do not set up filter_id column */
/*  10/26/01 gam Change measured from 1 to 0 when creating FFTPlan to remove warnings:
     LALMalloc warning: zero size allocation and LALFree warning: tried to free a freed pointer.
     These warning occur only when measure = 1 and the FFTPlan is for a power of 2 input data points.
     Note: use measure equal 0 for faster estimated plan, nonzero for more exact plan. */
/*  10/30/01 gam When params->generateTestData == 'y' handle case when params->vecInputData not yet created. */
/*  11/02/01 gam Replace int with INT4 in MoreTimes */
/*  11/02/01 gam Add testTypes 3 and 4
    0 = sinusoidal test signal only;
    1 = 0 + gaussian noise (mean = 0, std dev = 1);
    2 = add test signal to input data;
    3 = same as 0 but without phase or amplitude modulation due to Earth's motion;
    4 = same as 2 but without phase or amplitude modulation due to Earth's motion; */
/*  11/02/01 gam Add drop out percentage into dcHistory structure for each SFT. */
/*  11/05/01 gam Add more meta data into the dcHistory structure for each SFT. */
/*  11/05/01 gam Change filterID to dataCondDesc; this is a string with data conditioning info */
/*  11/05/01 gam add dsoName and percentPad as parameters; add these and dataConDesc to dcHistory structure for each SFT */
/*  11/05/01 gam Find power in freq bin closest to f0 = pwF0 and std def pwStdDev; temporarily set SNR = (pwF0 - pwMean)/pwStdDev. */
/*  11/05/01 gam Replace pwMaxOverpwMean with pwF0OverpwMean = pwF0/pwMean. Set Confidence = 1 - exp(-pwF0/pwMean)  */
/*  11/05/01 gam Replace phaseForPWMax with phaseForPWF0 */
/*  11/05/01 gam Replace freqPWMax with freqPWF0 */
/*  11/10/01 gam Change template and test RA, Dec, Freq, Freq derivatives, and polarization from REAL4 to REAL8. */
/*  11/10/01 gam Change database entries for RA, Dec, Freq from REAL4 to REAL8. */
/*  11/10/01 gam Change KNOWNPULSARDEMODH_LIGOBANDMIN to 0.0 Hz and KNOWNPULSARDEMODH_LIGOBANDMAX to 2048 Hz. */
/*  11/10/01 gam Add an SFTStats data structure; store max, mean, and stddev stats for each SFT in this structure; store values in SFT history */
/*  11/10/01 gam Add static function CalcDemodPowerStats to calculate SFTStats */
/*  11/10/01 gam Add parameter nDeFT =(params->if0Max-params->if0Min)*params->mCohSFT = Number of bins in a DeFT. */
/*  11/11/01 gam Add code to handle time delays introduced into by resampling and filtering. */
/*  11/17/01 gam Replace double with REAL8 in MoreTimes */
/*  11/17/01 gam Do not vet input stop time, since vetting inputN and deltaT is enough.  Stop time since it can be ambigous,
             e.g, can = start time + N*deltaT or start time + (N - 1)*deltaT. */
/*  11/17/01 gam Instead of putting stop time in the SFT, put number of input data point = nSFT. This is less ambiguous than the stop time. */
/*  11/17/01 gam Adjust order of meta data in SFT output history structure */
/*  11/18/01 gam Allow channel as an input parameter */
/*  11/27/01 gam Add code to pad data with mean value between cutoffs using a quality channel. */
/*  11/27/01 gam Initialize percentPad as -1, which mean percentPad was not set since dropout and quality was not handled */
/*  12/05/01 gam Add error checking iStartTime = index to gpsStartTime is not negative or past end of input data */
/*  12/05/01 gam Fix errors in checking for errors in declination, frequency derivatives, and mCohSFTPerCall */
/*  12/06/01 gam Updated basic documentation */
/*  12/13/01 gam Add column "search" to BuildOutput for SGNL_DPERIDIC database table */
/*  01/28/02 gam Change code to handle new definition of LALUnitPair. */
/*  02/07/02 gam Fix bug in quality channel code; cutoffs were incorrectly applied; disable use of cutoffs for now. */
/*  03/06/02 gam Save SFTs as COMPLEX8 rather than as COMPLEX16 data. */
/*  04/16/02 gam Complet code change to handle new definition of LALUnitPair. */
/*  05/13/02 gam Temporarily comment out INCLUDE_DEMOD_CODE */
/*  05/13/02 gam Uncomment INCLUDE_DEMOD_CODE to work on including new version of LALDemod */
/*  05/14/02 gam FFT data structure now holds COMPLEX8 frequency series; change xHat and SFTData to hold COMPLEX8 rather than COMPLEX16 data. */

/* TO DO LIST */
/* Improved quality channel code to better handle quality and cutoffs GW_CHANNEL_MAX_VALUE_EXPECTED and GW_CHANNEL_MIN_VALUE_EXPECTED */
/* Enhance MoreTimes to handle arbitary data drop outs and to zero pad missing data; rename to SetTimesHandleDropOuts */
/* Drop out includes data drop out, interferometer drop out of lock, and other data quality issues */
/*  Example data structure that gap_info is stored in:

                    <ilwd name='LDAS_History'>
                      <ilwd name='gap_info' size='5'>
                        <int_4u name='fill_method'>0</int_4u>
                        <int_4u dims='4' name='time_range'>688011799 999511718 688011801 0</int_4u>
                        <int_4u dims='4' name='time_range'>688011802 0 688011803 0</int_4u>
                        <int_4u dims='4' name='time_range'>688011804 0 688011805 0</int_4u>
                        <int_4u dims='4' name='time_range'>688011808 0 688011809 999999999</int_4u>
                      </ilwd>
                    </ilwd>

*/
/* Add amplitude modulation to testTypes 0 and 1 */

#define INCLUDE_DEMOD_CODE
#define INCLUDE_TEST_DATA_CODE

/* #define DEBUG_KNOWNPULSARDEMOD */
/* #define DEBUG_SFT_CODE */
/* #define DEBUG_OUTPUT_SFT_FILES */
/* #define DEBUG_DEMOD_CODE */
/* #define DEBUG_KNOWNPULSARDEMODOUTPUTFILES */
/* #define DEBUG_BUILDOUTPUT_CODE */
/* #define DEBUG_BUILDFRAMEOUTPUT_CODE */
/* #define DEBUG_TEST_CODE */
/* #define DEBUG_FRACREMAINING_CODE */
/* #define DEBUG_TIMEDELAY_CODE */
/* #define DEBUG_QUALITYCHANNEL_CODE */

#include <stdio.h>
#include <math.h>
#include <lal/LALStdlib.h>
#include <lal/AVFactories.h>
#include <lal/RealFFT.h>
#include <lal/LALConstants.h>
#include <lal/LALDemod.h>
#include <lal/Units.h>
#include <KnownPulsarDemod.h>

NRCSID( KNOWNPULSARDEMODC, "$Id: KnownPulsarDemod.c,v 1.11 2002/05/14 17:35:39 gmendell Exp $");

/********************** prototypes for local static function definitions ********************/
/* void tdb(REAL8 alpha, REAL8 delta, REAL8 t_GPS, REAL8 *T, REAL8 *Tdot, const CHAR *sw); */
void tdb(REAL8 alpha, REAL8 delta, REAL8 t_GPS, REAL8 *T, REAL8 *Tdot); /*  10/08/01 gam */
/* void times(REAL8 tSFT, int howMany, LIGOTimeGPS *ts, INT4 sw); */
/* void MoreTimes(REAL8 tSFT, int howMany, LIGOTimeGPS *ts, gpsTimeInterval *timeIntervals, INT4 sw); */ /* 10/10/01 gam */
/* void MoreTimes(REAL8 tSFT, int howMany, UINT4 gpsStartTime, LIGOTimeGPS *ts, gpsTimeInterval *timeIntervals, INT4 sw); */
void MoreTimes(REAL8 tSFT, INT4 howMany, UINT4 gpsStartTime, LIGOTimeGPS *ts, gpsTimeInterval *timeIntervals, INT4 sw); /* 11/02/01 gam */
/* 11/10/01 gam */ /* Calculate stats associated with frequency series */
/* void CalcDemodPowerStats( const COMPLEX16Vector *fsVec, DemodPowerStats *stats, REAL8 fStart, REAL8 df ); */ /* 05/14/02 gam */
void CalcDemodPowerStats( const COMPLEX8Vector *fsVec, DemodPowerStats *stats, REAL8 fStart, REAL8 df ); /* 05/14/02 gam */
/* 11/27/01 gam */  /* Calculate the mean value of a time series with cutoffs */
/* void CalcMeanWithCutoffs( const REAL4Vector *tInputVec, REAL4 cutoffMin, REAL4 cutoffMax, REAL4 *meanWithCuttoffs); */
/* 11/11/01 gam */ /* Find index of input data that corresponds to gpsStartTime */
void GetStartTimeIndex(const gpsTimeInterval *times, const UINT4 *gpsStartTime, UINT4 *iStartTime);
REAL8 ComputePhase(REAL8 x, REAL8 y);
static void BuildOutput(LALStatus*            const status,
	    		const DemodSearchParams* const params,
	    		LALSearchOutput*      const output);
/*
static void BuildFrameOutput(LALStatus* const status,
	    const DemodSearchParams*    const params,
	    FFT                         **SFTData,
            gpsTimeInterval             *timeIntervals,
	    LALSearchOutput*            const output);
*/
/* 11/10/01 gam */ /* Add in SFTStats */
static void BuildFrameOutput(LALStatus* const status,
	    const DemodSearchParams*    const params,
	    FFT                         **SFTData,
            DemodPowerStats             **SFTStats,
            gpsTimeInterval             *timeIntervals,
	    LALSearchOutput*            const output);

/* Initialization routine */
void
LALInitSearch(
    LALStatus             *status,
    void                 **searchParams,
    LALInitSearchParams   *initSearchParams
    )
{
  /* Declare variables */
  DemodSearchParams *params;
  UINT2 i;
  REAL8 totalDeltaT; /* 11/10/01 gam  Changed from REAL4 to REAL8 */
  REAL8 shiftedFreq; /* 11/10/01 gam  Changed from REAL4 to REAL8 */

  INITSTATUS( status, "LALInitSearch", KNOWNPULSARDEMODC );

  #ifdef DEBUG_KNOWNPULSARDEMOD
  	fprintf(stdout, "\nLALInitSearch\n");
  	fflush(stdout);
  #endif

  /* Attach DemodSearchParams to searchParams */
  *searchParams = params = LALMalloc(sizeof( DemodSearchParams ));

  /* ----------------- parse the filter parameter arguments --------------*/
  /* ----------------- and initialise other parameters --------------------*/

  #ifdef DEBUG_KNOWNPULSARDEMOD
  	printf("\nargc = %i\n",initSearchParams->argc);
  #endif

  /* Parameters passed as command line arguments */
  /*  BOOLEAN doDemod                   Demodulate the SFTs into DeFTs */
  /*  INT4    mObsCoh                   Number of DeFTs to compute from input observation time */
  /*  UINT4   gpsStartTime              GPS start time of data requested */
  /*  REAL8   deltaT;                   time step in seconds = 1/sample rate */
  /*  UINT4   inputN;                   number of input data points */
  /*  REAL8   tSFT;                     duration in seconds of SFTs (= nSFT*deltaT = Time between start of one SFT and next SFT for sequential SFTs) */
  /*  INT4    nSFT;                     number of time domain data points used to compute one SFTs */ /* 10/11/01 gam */
  /*  UINT4    mCohSFTPerCall;            number of SFTs to compute during each call to ApplySearch */ /* 10/11/01 gam */
  /*  REAL8   templateRA;               Right Ascension of template */
  /*  REAL8   templateDec;              Declination of the template */
  /*  REAL8   templatePhase0;           initial or midpoint phase of template */
  /*  REAL8   templateFreq0;            initial or midpoint frequency of template */
  /*  REAL8   templateFreq0Band;        Band around f0 to compute DeFT */
  /*  REAL8   templateFreqDeriv0;       initial or midpoint first deriv of freq of template */
  /*  REAL8   templateFreqSecondDeriv0; initial or midpoint second deriv of freq of template */
  /*  REAL8   templateFreqThirdDeriv0;  initial or midpoint third deriv of freq of template */
  /*  REAL8   templateFreqFourthDeriv0; initial or midpoint fourth deriv of freq of template */
  /*  REAL8   templateFreqFifthDeriv0;  initial or midpoint fifth deriv of freq of template */
  /*  CHAR    generateTestData;         Generate test data: y = yes, n = no. */
  /*  INT2    testType;                 0 = signal only; 1 = with noise; 2 = add test signal to input data */
  /*  REAL8   testRA;                   Right Ascension of test */
  /*  REAL8   testDec;                  Declination of the test */
  /*  REAL8   testAmplitude;            Amplitude of the test */
  /*  REAL8   testPhase0;               phase of test */
  /*  REAL8   testFreq0;                frequency of test */
  /*  REAL8   testFreqDeriv0;           first deriv of freq of test */
  /*  REAL8   testFreqSecondDeriv0;     second deriv of freq of test */
  /*  REAL8   testFreqThirdDeriv0;      third deriv of freq of test */
  /*  REAL8   testFreqFourthDeriv0;     fourth deriv of freq of test */
  /*  REAL8   testFreqFifthDeriv0;      fifth deriv freq of test */
  /*  REAL8   testPolarization;         Polarization of test; NOT YET IMPLEMENTED */
  /*  INT4    testDeletions;            If deletions > 1 then gaps of deletions*tSFT occur between tSFT segments of test data */
  /*  CHAR    *dataConDesc;             Information about data conditioning done  */ /* 10/16/01 gam */ /* 11/05/01 gam */
  /*  CHAR    *IFO;                     Identifies the interferometer name */ /* 10/16/01 gam */
  /*  CHAR    *targetName;              Identifies the target of the search */ /* 10/16/01 gam */
  /*  CHAR    channel[dbNameLimit];     Channel Name */ /* 11/18/01 gam */

 /* 10/16/01 gam */
 params->dataConDesc = NULL; /* 11/05/01 gam */
 params->IFO = NULL;
 params->targetName = NULL;

 /* Set up name of this DSO */
 params->dsoName = NULL; /* 11/05/01 gam */
 params->dsoName = (CHAR *) LALMalloc( (strlen( "KPD_DSO" ) + 1) * sizeof(CHAR) );
 strcpy(params->dsoName, "KPD_DSO");
 strcpy(params->channel, "NotSet"); /* 11/18/01 gam Initialize */

 for (i = 1; i < initSearchParams->argc; i++) {

	#ifdef DEBUG_KNOWNPULSARDEMOD
		/* printf("\nargv[%i] = %g\n",i,atof(initSearchParams->argv[i])); */
	#endif

	switch(i) {
		case  1: params->doDemod = (BOOLEAN)atol(initSearchParams->argv[i]); break;
		case  2: params->mObsCoh = (INT4)atol(initSearchParams->argv[i]); break;
		case  3: params->gpsStartTime = (UINT4)atol(initSearchParams->argv[i]); break;
		case  4: params->deltaT = (REAL8)atof(initSearchParams->argv[i]); break; /* 10/10/01 gam */
		case  5: params->inputN = (UINT4)atol(initSearchParams->argv[i]); break;
		case  6: params->tSFT = (REAL8)atof(initSearchParams->argv[i]); break; /* 10/10/01 gam */
		case  7: params->nSFT = (INT4)atol(initSearchParams->argv[i]); break; /* 10/11/01 gam */
		case  8: params->mCohSFTPerCall = (UINT4)atol(initSearchParams->argv[i]); break; /* 10/11/01 gam */
		case  9: params->templateRA = (REAL8)atof(initSearchParams->argv[i]); break;
		case 10: params->templateDec = (REAL8)atof(initSearchParams->argv[i]); break;
		case 11: params->templatePhase0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 12: params->templateFreq0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 13: params->templateFreq0Band = (REAL8)atof(initSearchParams->argv[i]); break;
		case 14: params->templateFreqDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 15: params->templateFreqSecondDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 16: params->templateFreqThirdDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 17: params->templateFreqFourthDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 18: params->templateFreqFifthDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 19: params->generateTestData = initSearchParams->argv[i][0]; break;
		case 20: params->testType = (INT2)atoi(initSearchParams->argv[i]); break;
		case 21: params->testRA = (REAL8)atof(initSearchParams->argv[i]); break;
		case 22: params->testDec = (REAL8)atof(initSearchParams->argv[i]); break;
		case 23: params->testAmplitude = (REAL8)atof(initSearchParams->argv[i]); break;
		case 24: params->testPhase0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 25: params->testFreq0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 26: params->testFreqDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 27: params->testFreqSecondDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 28: params->testFreqThirdDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 29: params->testFreqFourthDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 30: params->testFreqFifthDeriv0 = (REAL8)atof(initSearchParams->argv[i]); break;
		case 31: params->testPolarization = (REAL8)atof(initSearchParams->argv[i]); break;
		case 32: params->testDeletions = (INT4)atol(initSearchParams->argv[i]); break;
                /* 11/05/01 gam */
		case 33: params->dataConDesc = (CHAR *) LALMalloc( (strlen( initSearchParams->argv[i] ) + 1) * sizeof(CHAR) );
		         strcpy(params->dataConDesc, initSearchParams->argv[i]); break;
		case 34: params->IFO = (CHAR *) LALMalloc( (strlen( initSearchParams->argv[i] ) + 1) * sizeof(CHAR) );
		         strcpy(params->IFO, initSearchParams->argv[i]); break;
		case 35: params->targetName = (CHAR *) LALMalloc( (strlen( initSearchParams->argv[i] ) + 1) * sizeof(CHAR) );
		         strcpy(params->targetName, initSearchParams->argv[i]); break;
		case 36: strcpy(params->channel, initSearchParams->argv[i]); break;
	}
  } /* end for (i=1; i<argc; i++)  */

  #ifdef DEBUG_TEST_CODE
  	fprintf(stdout, "\nparams->generateTestData = %c\n",params->generateTestData);
  	fflush(stdout);
  #endif

  params->searchMaster = initSearchParams->nodeClass;
  params->rank         = 0;
  params->numSlaves    = 0;
  params->curSlaves    = 0;
  /* strcpy(params->channel, ""); */ /* 11/18/01 gam Moved above */
  params->vecInputData  = NULL;
  params->freqPWF0 = 0.0;       /* freq detected with maximum power */
  params->pwMax = 0.0;          /* power = absolute square of the amplitude at this freq */
  params->phaseForPWF0 = 0.0;   /* phase that goes with amplitude at max power */
  params->pwMean = 0.0;         /* mean power detected (mean computed less max power) */
  params->pwF0OverpwMean = 0.0; /* ratio of powMax to pwMean */ /* 11/05/01 gam */
  params->pwF0 = 0.0;           /* Power in the bin closest to templateFreq0 */   /* 11/05/01 gam */
  params->pwStdDev = 0.0;       /* Standard deviation in power (less pwF0) */ /* 11/05/01 gam */

  /* params->percentPad = 0.0; */ /* 11/27/01 gam */  /* Percent of input time series to produce each SFT was zero padded to fill in gaps */ /* 11/05/01 gam */
  params->percentPad = -0.01; /* 11/27/01 gam  will = -1% means not set*/ /* Percent of input time series to produce each SFT was zero padded to fill in gaps */ /* 11/05/01 gam */

  /* parameters for keeping track of which SFTs to do, and if all SFTs are done; used with mCohSFTPerCall */
  params->firstK = 0;                 /* Index of the first SFT to compute and store in the SFTData structure this call to ApplySearch */
  params->lastK = 0;                  /* Index of the last SFT to compute and store in the SFTData Structure this call to ApplySearch */
  params->finishedSFTs = 0;   /* Set equal to true when all SFTS for this job have been computed */

  params->ifMin = 0;      /* Index of minimum frequency in SFT band */
  params->fMin = 0.0;     /* min freq in SFT band */
  params->ifMax = 0;      /* Index of maximum frequency in SFT band */
  params->if0Min = 0;     /* Min search freq index in coarse-grained frequency resolution */
  params->f0Min = 0.0;    /* Min freq in coarse-grained freq resolution */
  params->if0Max = 0;     /* Max search freq index in coarse-grained frequency resolution */
  params->f0Max = 0.0;    /* Max freq in coarse-grained freq resolution */
  params->nDeFT = 0;     /* Number of bins in a DeFT */ /* 11/10/01 gam */

  if ( strstr(params->channel, "NotSet") ) {
      strcpy(params->channel, "H2\\:LSC-AS_Q"); /* This is the default */
  }

  /* Validate parameters */

  if (params->mObsCoh != 1) {
    ABORT( status, KNOWNPULSARDEMODH_EMOBSCOH, KNOWNPULSARDEMODH_MSGEMOBSCOH);
  }
  if (params->deltaT <= 0.0) {
    ABORT( status, KNOWNPULSARDEMODH_EDELTAT, KNOWNPULSARDEMODH_MSGEDELTAT);
  }
  if (params->inputN <= 0) {
    ABORT( status, KNOWNPULSARDEMODH_EINPUTN, KNOWNPULSARDEMODH_MSGEINPUTN);
  }

  /* if (params->tSFT <= 0.0 || params->tSFT > params->deltaT*((REAL4)(params->inputN - 1))) */
  /* 10/10/01 gam */
  /* if (params->tSFT <= 0.0 || params->tSFT > params->deltaT*((REAL8)(params->inputN - 1))) */
  /* 10/11/01 gam */
  if (params->tSFT <= 0.0 || params->tSFT > params->deltaT*((REAL8)params->inputN)) {
    ABORT( status, KNOWNPULSARDEMODH_ETSFT, KNOWNPULSARDEMODH_MSGETSFT);
  }

  /* 10/11/01 gam */
  if (params->nSFT <= 0 || params->nSFT > (INT4)params->inputN) {
    ABORT( status, KNOWNPULSARDEMODH_ENSFT, KNOWNPULSARDEMODH_MSGENSFT);
  }

  /* 10/11/01 gam */
  /* if ( params->tSFT != params->deltaT*(REAL8)params->nSFT ) */
  if (fabs((REAL8)((REAL8)params->tSFT - (REAL8)(params->deltaT*(REAL8)params->nSFT))) > 1.e-9*params->tSFT ) {
    ABORT( status, KNOWNPULSARDEMODH_ENTSFT, KNOWNPULSARDEMODH_MSGENTSFT);
  }
  /* Recompute tSFT to make it consistent with nSFT to double precision */
  params->tSFT = (REAL8)(((REAL8)params->deltaT)*((REAL8)params->nSFT));
  /* Check again! */
  if (fabs((REAL8)((REAL8)params->tSFT - (REAL8)(params->deltaT*(REAL8)params->nSFT))) > 1.e-9*params->tSFT ) {
    ABORT( status, KNOWNPULSARDEMODH_ENTSFT, KNOWNPULSARDEMODH_MSGENTSFT);
  }

  /* 10/11/01 gam */
  /* if (params->mCohSFTPerCall < 0) {} */
  /* 12/05/01 gam */
  if (params->mCohSFTPerCall <= 0) {
    ABORT( status, KNOWNPULSARDEMODH_EMCOHSFTPERCALL, KNOWNPULSARDEMODH_MSGEMCOHSFTPERCALL);
  }

  if (params->templateRA < 0 || params->templateRA > (REAL8)LAL_TWOPI) {
    /* Could be OK, but for now do not allow */
    ABORT( status, KNOWNPULSARDEMODH_ERA, KNOWNPULSARDEMODH_MSGERA);
  }

  /* 12/05/01 gam */
  /* if (params->templateDec < -1.0*(REAL8)LAL_PI_2 || params->templateRA > (REAL8)LAL_PI_2) {} */
  if (params->templateDec < -1.0*(REAL8)LAL_PI_2 || params->templateDec > (REAL8)LAL_PI_2) {
    ABORT( status, KNOWNPULSARDEMODH_EDEC, KNOWNPULSARDEMODH_MSGEDEC);
  }

  if (params->templateFreq0 - 0.5*params->templateFreq0Band < (REAL8)KNOWNPULSARDEMODH_LIGOBANDMIN ||
       params->templateFreq0 + 0.5*params->templateFreq0Band > (REAL8)KNOWNPULSARDEMODH_LIGOBANDMAX ||
       params->templateFreq0 + 0.5*params->templateFreq0Band > 0.5/params->deltaT) {

       ABORT( status, KNOWNPULSARDEMODH_EFREQ, KNOWNPULSARDEMODH_MSGEFREQ);
  }

  /* Estimate maximum shift in frequency during demodulation */

  /* totalDeltaT = params->deltaT*((REAL4)(params->inputN - 1)); */
  /* totalDeltaT = params->deltaT*((REAL8)(params->inputN - 1)); */ /* 10/10/01 gam */
  totalDeltaT = params->deltaT*((REAL8)params->inputN); /* 10/11/01 gam */
  shiftedFreq = params->templateFreq0;
  for (i = 1; i <= 5; i++) {
     /* shiftedFreq += ((REAL4)atof(initSearchParams->argv[23+i]))*pow(totalDeltaT,(REAL4)i); */
     /* shiftedFreq += ((REAL8)atof(initSearchParams->argv[25+i]))*pow(totalDeltaT,(REAL8)i); */ /* 10/11/01 gam */
     shiftedFreq += ((REAL8)atof(initSearchParams->argv[13+i]))*pow(totalDeltaT,(REAL8)i); /* 12/05/01 gam */
  }

  if  (shiftedFreq - 0.5*params->templateFreq0Band < (REAL8)KNOWNPULSARDEMODH_LIGOBANDMIN ||
       shiftedFreq + 0.5*params->templateFreq0Band > (REAL8)KNOWNPULSARDEMODH_LIGOBANDMAX ||
       shiftedFreq + 0.5*params->templateFreq0Band > 0.5/params->deltaT) {

       ABORT( status, KNOWNPULSARDEMODH_EFREQDERIV, KNOWNPULSARDEMODH_MSGEFREQDERIV);
  }

  if (params->generateTestData != 'y' && params->generateTestData != 'n') {
     ABORT( status, KNOWNPULSARDEMODH_EGENTEST, KNOWNPULSARDEMODH_MSGEGENTEST);
  }

  /* if (params->testType != 0 && params->testType != 1 && params->testType != 2) {} */ /* 11/02/01 gam */
  /* Note: testType is UINT2; cannot be negative */
  /* "Invalid test data type" */
  if (params->testType > 4) {
     ABORT( status, KNOWNPULSARDEMODH_ETESTTYPE, KNOWNPULSARDEMODH_MSGETESTTYPE);
  }

  /* 10/16/01 gam */

  /* "Invalid or null dataCondDesc" */ /* 11/05/01 gam */
  if ( !(params->dataConDesc) ){
    ABORT( status, KNOWNPULSARDEMODH_EDCONDESC, KNOWNPULSARDEMODH_MSGEDCONDESC);
  }

  /* "Invalid or null IFO" */
  if ( !(params->IFO) ){
    ABORT( status, KNOWNPULSARDEMODH_EIFO, KNOWNPULSARDEMODH_MSGEIFO);
  }

  /* "Invalid or null Target Name" */
  if ( !(params->targetName) ){
    ABORT( status, KNOWNPULSARDEMODH_ETARGETNAME, KNOWNPULSARDEMODH_MSGETARGETNAME);
  }

  /* "Invalid or null channel" */
  if ( strstr(params->channel, "NotSet") ){
    ABORT( status, KNOWNPULSARDEMODH_ECHANNEL, KNOWNPULSARDEMODH_MSGECHANNEL);
  }

  RETURN( status );

}

/* ConditionData routine */
void
LALConditionData(
    LALStatus             *status,
    LALSearchInput        *inout,
    void                  *searchParams,
    LALMPIParams	  *mpiParams
    )
{

 DemodSearchParams* const params = (DemodSearchParams *) searchParams;

 INITSTATUS( status, "LALConditionData", KNOWNPULSARDEMODC );
 ATTATCHSTATUSPTR (status);

 /* 10/12/01 gam */
 MPI_Comm_rank( *(mpiParams->comm), &params->rank );

 /* !!!!!!!!!!!!!!!!!!! TO START WITH, WILL RUN CODE ONLY ON THE SEARCH MASTER !!!!!!!!!!!!!!!!! */
 /* !!!!!!! LATER, WILL RUN ON THE SLAVES, AND ADD COMMUNICATION OF RESULTS BACK TO MASTER !!!!! */
 if (params->searchMaster) {

  BOOLEAN dataFound = 0;

  /* BOOLEAN gapsFound = 0; */ /* FOR FUTURE USE: Found gap_info in the input mdd history record  */ /* 11/27/01 gam */
  BOOLEAN qc1Found = 0;  /* Found quality channel qc1 */ /* 11/27/01 gam */
  UINT4 qc1Index = 0;    /* The quality channel was found in inout->sequences[qc1Index].data.real4 */ /* 11/27/01 gam */

  UINT4 i = 0;      /* all purpose index */
  UINT4 k = 0;      /* another all purpose index */
  UINT4 nData = 0;  /* 10/10/01 gam */
  UINT4 iStartTime = 0; /* 11/11/01 gam */

  /* Variables used to compute SFTs that are only needed in ConditionData */
  /* const INT4 measure = 1; */ /* Use 0 for faster estimated plan, nonzero for more exact plan. Used in new version LALCreate...FFTPlan() */
  const INT4 measure = 0;  /* 10/26/01 gam */ /* Use 0 for faster estimated plan, nonzero for more exact plan. Used in new version LALCreate...FFTPlan() */
  INT4 deletions=1;        /* Used when computing timeStamps.  For now hard code with default value.  BUT RESET IF GENERATING TEST DATA */

  for (i = 0; i < inout->numberSequences; ++i)
  {
    /* if (strstr(inout->sequences[i].name, "H2\\:LSC-AS_Q")) */ /* 11/18/01 gam */
    if (strstr(inout->sequences[i].name, params->channel))
    {

      params->times = inout->sequences[i].range.dTime;
      /* 11/11/01 gam */ /* Find index of input data that corresponds to gpsStartTime */
      GetStartTimeIndex(&params->times, &params->gpsStartTime, &iStartTime);
      #ifdef DEBUG_TIMEDELAY_CODE
  	fprintf(stdout, "\nFound channel, iStartTime = %i \n",iStartTime);
  	fflush(stdout);
      #endif
      /* 12/05/01 gam */
      if (iStartTime < 0 || iStartTime >= inout->sequences[i].dimensions[0]) {
          ABORT( status, KNOWNPULSARDEMODH_EISTARTTIME, KNOWNPULSARDEMODH_MSGEISTARTTIME);
      }

      /* const UINT4 nData = inout->sequences[i].dimensions[0]; */
      /* nData = inout->sequences[i].dimensions[0]; */ /* 10/10/01 gam */
      /* 11/11/01 gam */
      nData = inout->sequences[i].dimensions[0] - iStartTime;
      params->times.startSec = params->gpsStartTime;
      params->times.startNan = 0;

      LALSCreateVector(status->statusPtr, &(params->vecInputData), nData);

      /*
      memcpy(params->vecInputData->data, inout->sequences[i].data.real4,
	     sizeof(REAL4)*nData);
      */

      memcpy(params->vecInputData->data, (inout->sequences[i].data.real4 + iStartTime),
	     sizeof(REAL4)*nData);

      /* strcpy(params->channel, "H2:LSC-AS_Q"); */ /* 11/18/01 gam */

      params->unitSFT = lalADCCountUnit;  /* 10/12/01 gam */ /* HARD CODE FOR NOW; BUT SHOULD TRY SOMETHING LIKE CODE BELOW */
      /* const CHARVector *inputUnitString  = inout->sequences[i].units; */
      /* LALParseUnitString(status->statusPtr,&params->unitSFT,inputUnitString); */
      /* CHECKSTATUSPTR (status); */

      dataFound = 1;
      /* break; */ /* 11/27/01 gam */
      if (qc1Found && dataFound) {
      	break;
      }
    }
    else if (strstr(inout->sequences[i].name, "ifodmro"))
    {
      /* const INT2* dummyI2 = inout->sequences[i].data.int2s; */ /* 11/11/01 gam */
      INT2* dummyI2 = inout->sequences[i].data.int2s;

      params->times = inout->sequences[i].range.dTime;
      /* 11/11/01 gam */ /* Find index of input data that corresponds to gpsStartTime */
      GetStartTimeIndex(&params->times, &params->gpsStartTime, &iStartTime);
      #ifdef DEBUG_TIMEDELAY_CODE
  	fprintf(stdout, "\nFound ifodmro, iStartTime = %i \n",iStartTime);
  	fflush(stdout);
      #endif
      /* 12/05/01 gam */
      if (iStartTime < 0 || iStartTime >= inout->sequences[i].dimensions[0]) {
          ABORT( status, KNOWNPULSARDEMODH_EISTARTTIME, KNOWNPULSARDEMODH_MSGEISTARTTIME);
      }

      /* UINT4 k = 0; */ /* 10/12/01 gam */

      /* const UINT4 nData = inout->sequences[i].dimensions[0]; */
      /* nData = inout->sequences[i].dimensions[0]; */ /* 10/10/01 gam */

      /* 11/11/01 gam */
      nData = inout->sequences[i].dimensions[0] - iStartTime;
      params->times.startSec = params->gpsStartTime;
      params->times.startNan = 0;

      /*
       * All we want to do is convert the IFODMRO data from int to
       * float and get it to the ApplySearch routine
       */

      LALSCreateVector(status->statusPtr, &(params->vecInputData), nData);

      /*
       * Copy from int2s to real4
       */
      *dummyI2 = *dummyI2 + iStartTime; /* 11/11/01 gam */
      for (k = 0; k < params->vecInputData->length; ++k)
      {
	params->vecInputData->data[k] = (REAL4) *dummyI2++;
      }

      strcpy(params->channel, "ifodmro");

      params->unitSFT = lalVoltUnit;  /* 10/12/01 gam */ /* HARD CODE FOR NOW; BUT SHOULD TRY SOMETHING LIKE CODE BELOW */
      /* const CHARVector *inputUnitString  = inout->sequences[i].units; */
      /* LALParseUnitString(status->statusPtr,&params->unitSFT,inputUnitString); */
      /* CHECKSTATUSPTR (status); */

      dataFound = 1;
      /* break; */ /* 11/27/01 gam */
      if (qc1Found && dataFound) {
      	break;
      }
    }
    else if (strstr(inout->sequences[i].name, "qc1"))
    {
      /* 11/27/01 */ /* Found the quality channel qc1 */

      qc1Index = i;
      qc1Found = 1;
      if (qc1Found && dataFound) {
      	break;
      }


    } /* END if (strstr(inout->sequences[i].name, "H2\\:LSC-AS_Q")) elseif ... */
   }  /* END for (i = 0; i < inout->numberSequences; ++i) */

    /* 10/12/01 gam */
    /* Finish initializing variables used to compute SFTs */

        /* IMPORTANT, note that I define nSFT, which is related to nDeltaF in LALDeModTest */
        /* IMPORTANT, note that I define nDeltaF = nSFT/2 + 1 */   /* 09/28/01 gam */
        /* nSFT is the number of time domain data points in the input to LALFORWARDREAL and nDeltaF is the number of output data points */
        /* IMPORTANT, note that I assume deltaT is the same as dtEFF in LALDeModTest */

         /* params->pfwd = NULL; */     /* Plan for computing SFTs */
         params->pfwdReal = NULL;       /* Plan for computing FFTs from real input. */ /* 09/28/01 gam */

         params->tCoh = params->deltaT*((REAL8)params->inputN); /* Duration of DeFTs.  For now, this is taken as the time input into this code */

         params->dfSFT = 1.0/params->tSFT;                        /* Freq resolution of SFT = 1/duration of SFTs  */

         params->mCohSFT = floor(params->tCoh/params->tSFT);      /* Number of SFTs in one DeFT. */
         params->mObsSFT = params->mCohSFT*params->mObsCoh;       /* Total number of SFTs in the observation time. */

         params->nDeltaF = params->nSFT/2 + 1;                            /* Number of output data points in one SFT */ /* 09/28/01 gam */

         params->tvec = NULL;            /* Temporary holder in input when computing each SFT */ /* 09/28/01 gam */
         params->fvec = NULL;         /* Temporary holder of output when computing each SFT */

    /* 10/12/01 gam */
    /* Finish allocating memory and computing parameters that only get computed once when computing SFTs */

        /***** START CALCULATE USEFUL QUANTITIES BASED ON LALDemodTest.c *****/

	  params->ifMin = 0;
	  /* params->ifMax = params->nSFT - 1; */
	  params->ifMax = params->nDeltaF - 1; /* 09/28/01 gam */

	  params->fMin=params->dfSFT*params->ifMin;                          /* Minimum frequency for which search is conducted */
	  params->if0Max = ceil((params->templateFreq0+params->templateFreq0Band/2.0)/params->dfSFT);      /* Index of maximum frequency to include in DeFT */
	  params->f0Max=params->dfSFT*params->if0Max;                        /* Max freq in DeFT */
          params->if0Min = floor((params->templateFreq0-params->templateFreq0Band/2.0)/params->dfSFT);     /* Index of mimimum frequency to include in DeFT */
	  params->f0Min=params->dfSFT*params->if0Min;                        /* Min freq in DeFT */

	  params->nDeFT = (params->if0Max-params->if0Min)*params->mCohSFT; /* Number of bins in a DeFT. */ /* 11/10/01 gam */

	  params->funcName=tdb;    /* For now hard codes as is LALDemodTest */

 	  /* params->inputM = params->mObsSFT; */ /* 10/12/01 gam */ /* number of input data segments */ /* TO DO: Check that inputM is set up correctly */

         #ifdef DEBUG_DEMOD_CODE
          fprintf(stdout,"\n\nifMin = %i, ifMax = %i \n",params->ifMin, params->ifMax);
          fprintf(stdout,"if0Min = %i, if0Max = %i \n",params->if0Min,params->if0Max);
          fprintf(stdout,"f0Min = %g, f0Max = %g \n",params->f0Min, params->f0Max);
          /* fprintf(stdout,"nSFT = %i, tSFT = %g, dfSFT = %g, f0Band = %g \n",params->nSFT,params->tSFT,params->dfSFT,params->templateFreq0Band); */
          fprintf(stdout,"nSFT = %i, nDeltaF = %i, tSFT = %g, dfSFT = %g, f0Band = %g \n",params->nSFT,params->nDeltaF,params->tSFT,params->dfSFT,params->templateFreq0Band); /* 09/28/01 gam */
          fprintf(stdout,"mCohSFT = %i \n",params->mCohSFT);
          fprintf(stdout,"mObsSFT = %i, mObsCoh = %i \n",params->mObsSFT,params->mObsCoh);
          fprintf(stdout,"params->nDeFT %i \n\n",params->nDeFT);
          fflush(stdout);
         #endif

       /***** END CALCULATE USEFUL QUANTITIES BASED ON LALDemodTest.c *****/

       /***** START CALL ROUTINE TO GENERATE TIMESTAMPS BASED ON LALDemodTest.c *****/

	/* params->timeIntervals =(gpsTimeInterval *)LALMalloc(params->mObsSFT*sizeof(gpsTimeInterval)); */
	params->timeIntervals =(gpsTimeInterval *)LALCalloc(params->mObsSFT, sizeof(gpsTimeInterval));

	params->timeStamps=(LIGOTimeGPS *)LALMalloc(params->mObsSFT*sizeof(LIGOTimeGPS));
	MoreTimes(params->tSFT, params->mObsSFT, params->gpsStartTime, params->timeStamps, params->timeIntervals, deletions); /* 10/10/01 gam */

      /***** END CALL ROUTINE TO GENERATE TIMESTAMPS BASED ON LALDemodTest.c *****/

        /* Create vector to hold frequency series */
    	LALCCreateVector(status->statusPtr, &params->fvec, (UINT4)params->nDeltaF); /* 09/28/01 gam */

        /* Create vector to hold time series */
    	LALSCreateVector(status->statusPtr, &params->tvec, (UINT4)params->nSFT);

        /* Allocate memory for the SFTData structure */
    	params->SFTData=(FFT **)LALMalloc(params->mObsSFT*sizeof(FFT *));
    	for(i=0;i<params->mObsSFT;i++)
    	{
		params->SFTData[i]=(FFT*)LALMalloc(sizeof(FFT));

		/* 05/14/02 gam */
		/* params->SFTData[i]->fft=(COMPLEX16FrequencySeries *)LALMalloc(sizeof(COMPLEX16FrequencySeries));
		params->SFTData[i]->fft->data=(COMPLEX16Vector *)LALMalloc(sizeof(COMPLEX16Vector)); */
		params->SFTData[i]->fft=(COMPLEX8FrequencySeries *)LALMalloc(sizeof(COMPLEX8FrequencySeries));
		params->SFTData[i]->fft->data=(COMPLEX8Vector *)LALMalloc(sizeof(COMPLEX8Vector));

		/* params->SFTData[i]->fft->data->data=(COMPLEX16
			*)LALMalloc(params->nSFT*sizeof(COMPLEX16)); */
		/* params->SFTData[i]->fft->data->data=(COMPLEX16 *)LALMalloc(params->nDeltaF*sizeof(COMPLEX16)); */ /* 09/28/01 gam */ /* 05/14/02 gam */
		params->SFTData[i]->fft->data->data=(COMPLEX8 *)LALMalloc(params->nDeltaF*sizeof(COMPLEX8)); /* 05/14/02 gam */
    	} /* END for(i=0;i<params->mObsSFT;i++) */

        /* 11/10/01 gam */ /* Allocate memory for the SFTStats structure */
    	params->SFTStats=(DemodPowerStats **)LALMalloc(params->mObsSFT*sizeof(DemodPowerStats *));
    	for(i=0;i<params->mObsSFT;i++)
    	{
		params->SFTStats[i]=(DemodPowerStats*)LALMalloc(sizeof(DemodPowerStats));
    	} /* END for(i=0;i<params->mObsSFT;i++) */

      /* Compute plan for FFTW */
      /* LALCreateForwardComplexFFTPlan(status->statusPtr, &params->pfwd, (UINT4)params->nSFT, measure); */
      LALCreateForwardRealFFTPlan( status->statusPtr, &params->pfwdReal, (UINT4)params->nSFT, measure ); /* 09/28/01 gam */

      /* 10/15/01 gam */ /* Moved here */
      #ifdef INCLUDE_TEST_DATA_CODE
         if (params->generateTestData == 'y') {

	     /* variable used to generate test data */
	     REAL8 ts=0.0;
	     REAL8 tn=0.0;
	     REAL8 ts0=0.0;
	     REAL4 temp = 0.0;
	     REAL4Vector *noise = NULL;
	     static RandomParams *rndParams;
	     INT4 seed=0;
	     REAL8 baseTbary0, tDot, t, tBary;
	     REAL8 dTbary, dTbary2, dTbary3, dTbary4, dTbary5, dTbary6;

            /* IF ABOVE TRUE, USE TEST DATA OR ADD TEST DATA TO INPUT DATA */

            #ifdef DEBUG_TEST_CODE
  		 fprintf(stdout, "\nGENERATING TEST DATA FOR params->testType = %i\n",params->testType);
  		 fflush(stdout);
            #endif

	    /* Regenerate time stamps to allow gaps in test data */
	    deletions =	params->testDeletions;
	    /* MoreTimes(params->tSFT, params->mObsSFT, params->timeStamps, params->timeIntervals, deletions); */ /* 10/10/01 gam */
	    MoreTimes(params->tSFT, params->mObsSFT, params->gpsStartTime, params->timeStamps, params->timeIntervals, deletions);

	     /* conversion: detector->barycenter time,  t=GPS_0 */
	    ts0=(REAL8)(params->timeStamps[0].gpsSeconds) + (REAL8)(params->timeStamps[0].gpsNanoSeconds)*1.0E-9;
	    /* (*params->funcName)((REAL8)params->testRA,(REAL8)params->testDec, ts0, &baseTbary0, &tDot, modulation); */
	    (*params->funcName)((REAL8)params->testRA,(REAL8)params->testDec, ts0, &baseTbary0, &tDot); /*  10/08/01 gam */

	    /* if (params->inputN != params->vecInputData->length) {} */ /* 10/30/01 gam */
	    if (params->inputN != nData) {
	    	if (params->testType != 2) {
		        /* Need to resize vecInputData */
			/* 10/30/01 gam */
			if (dataFound) {
	    			LALSDestroyVector(status->statusPtr, &(params->vecInputData));
			}
      			LALSCreateVector(status->statusPtr, &(params->vecInputData),params->inputN);
		} else {
			/* ERROR TEST DATA LENGTH MUST EQUAL INPUT DATA LENGTH FOR CASE 2; ADD ABORT HERE */
			/* "Invalid input data or inputN for test type 2" */
    			ABORT( status, KNOWNPULSARDEMODH_EDATATESTTWO, KNOWNPULSARDEMODH_MSGEDATATESTTWO); /* 10/30/01 gam */
		}
	    }

    	    LALSCreateVector(status->statusPtr, &noise, params->inputN);
	    for(i=0;i<params->inputN;i++) {
	       noise->data[i]=0.0;  /* initialize */
	    }

	    /* if (params->testType == 1 || params->testType == 4) {} */ /* 11/02/01 gam */
	    if (params->testType == 1 || params->testType == 4) {
	       /* fill noise vector with normal deviates */
	       LALCreateRandomParams(status->statusPtr, &rndParams, seed);
	       LALNormalDeviates(status->statusPtr, noise, rndParams);
	       LALDestroyRandomParams(status->statusPtr, &rndParams);
	    }

	    for(k=0; k<params->mObsSFT; k++)
	    {
		ts=(REAL8)(params->timeStamps[k].gpsSeconds)*1.00;
		tn=(REAL8)(params->timeStamps[k].gpsNanoSeconds)*1.00E-9;

		for(i=0; i < params->nSFT; i++)
		{
			t=ts+tn+(REAL8)i*params->deltaT;

			/* conversion again */
			/* (*params->funcName)((REAL8)params->testRA, (REAL8)params->testDec, (REAL8)t, &tBary, &tDot, modulation); */
			(*params->funcName)((REAL8)params->testRA, (REAL8)params->testDec, (REAL8)t, &tBary, &tDot); /*  10/08/01 gam */

	    		if (params->testType <= 2) {
			   dTbary=tBary-baseTbary0;
			} else {
			   dTbary=t-ts0;  /* Do not amplitude or phase modulate data due to Earth's motion */ /* 11/02/01 gam */
			}
			/* pipeline for better performance */
			dTbary2=dTbary*dTbary;
			dTbary3=dTbary2*dTbary;
			dTbary4=dTbary3*dTbary;
			dTbary5=dTbary4*dTbary;
			dTbary6=dTbary5*dTbary;

			/* create time series according to test phase model */
			temp = params->testAmplitude*cos(LAL_TWOPI*(
				  params->testPhase0 +
			          params->testFreq0*dTbary +
				  0.5*params->testFreqDeriv0*dTbary2 +
				  (1.0/3.0)*params->testFreqSecondDeriv0*dTbary3 +
				  0.25*params->testFreqThirdDeriv0*dTbary4 +
				  0.20*params->testFreqFourthDeriv0*dTbary5 +
				  (1.0/6.0)*params->testFreqFifthDeriv0*dTbary6
				));
	    		/* if (params->testType == 0) {} */ /* 11/02/01 gam */
	    		if (params->testType == 0 || params->testType == 3) {
				params->vecInputData->data[i + k*params->nSFT] = temp;
	    		        /* {} else if (params->testType == 1) {} */ /* 11/02/01 gam */
	    		} else if (params->testType == 1 || params->testType == 4) {
				params->vecInputData->data[i + k*params->nSFT] = temp + noise->data[k*params->nSFT + i];
			} else if (params->testType == 2) {
				params->vecInputData->data[i + k*params->nSFT] = params->vecInputData->data[i + k*params->nSFT] + temp;
			}
		} /* end for(i=0; i < params->nSFT; i++) */
	     }	/* end for(k=0; k<params->mObsSFT; k++) */
	     LALDestroyVector(status->statusPtr, &noise);

             nData = params->inputN;
             /* strcpy(params->channel, "test"); */ /* 11/05/01 gam */
	     switch(params->testType) {
			case 0: strcpy(params->channel, "test0"); break;
			case 1: strcpy(params->channel, "test1"); break;
			case 2: strcpy(params->channel, "test2"); break;
			case 3: strcpy(params->channel, "test3"); break;
			case 4: strcpy(params->channel, "test4"); break;
	     }
	     params->times.timeStepSize = params->deltaT;
	     params->times.startSec = params->timeIntervals[0].startSec;
	     params->times.startNan = params->timeIntervals[0].startNan;
	     params->times.stopSec = params->timeIntervals[params->mObsSFT - 1].stopSec;
	     params->times.stopNan = params->timeIntervals[params->mObsSFT -1].stopNan;
             params->unitSFT = lalVoltUnit;  /* STILL NEED TO DECIDE ON UNIT FOR TEST DATA */
             dataFound = 1;

         } /* END if (params->generateTestData == 'y') */
   #endif

   /* 10/10/01 gam */
   /* Check that data requested was found (if not generating test data) */
   if (params->generateTestData == 'n') {

	/* "No input data was found" */
        if (!dataFound) {
    		ABORT( status, KNOWNPULSARDEMODH_ENODATA, KNOWNPULSARDEMODH_MSGENODATA);
	}

        #ifdef DEBUG_KNOWNPULSARDEMOD
               fprintf(stdout,"\nnData = %i \n", nData);
               fflush(stdout);
        #endif

	/* "Invalid number of input data points" */
	if (params->inputN != nData) {
    		ABORT( status, KNOWNPULSARDEMODH_ENDATA, KNOWNPULSARDEMODH_MSGENDATA);
	}

        #ifdef DEBUG_KNOWNPULSARDEMOD
               fprintf(stdout,"\ntimeStepSize = %g \n", params->times.timeStepSize);
               fflush(stdout);
        #endif

	/* "Incorrect input data time step" */
	if (params->times.timeStepSize != params->deltaT) {
    		ABORT( status, KNOWNPULSARDEMODH_ETIMESTEP, KNOWNPULSARDEMODH_MSGETIMESTEP);
	}

        #ifdef DEBUG_KNOWNPULSARDEMOD
               fprintf(stdout,"\ntimes.startSec = %i \n", params->times.startSec);
               fprintf(stdout,"times.startNan = %i \n", params->times.startNan);
               fflush(stdout);
        #endif

	/* "Incorrect input data start time" */
	if (params->gpsStartTime != params->times.startSec || params->times.startNan != 0) {
    		ABORT( status, KNOWNPULSARDEMODH_ESTARTTIME, KNOWNPULSARDEMODH_MSGESTARTTIME);
	}

        #ifdef DEBUG_KNOWNPULSARDEMOD
               /* fprintf(stdout,"\nComputed stop time = %f \n",(REAL8)((REAL8)params->times.startSec + params->times.timeStepSize*(REAL8)(nData - 1))); */
 	       /* fprintf(stdout,"Input stop time    = %f \n",(REAL8)((REAL8)params->times.stopSec + (REAL8)1.0e-9*(REAL8)params->times.stopNan)); */
               /* fprintf(stdout,"Difference         = %g \n", (REAL8)((REAL8)params->times.stopSec + (REAL8)1.0e-9*(REAL8)params->times.stopNan)
	                 -  (REAL8)((REAL8)params->times.startSec + params->times.timeStepSize*(REAL8)(nData - 1)) ); */
	       /* 10/11/01 gam */
               fprintf(stdout,"\nComputed stop time = %f \n",(REAL8)((REAL8)params->times.startSec + params->times.timeStepSize*(REAL8)nData));
 	       fprintf(stdout,"Input stop time    = %f \n",(REAL8)((REAL8)params->times.stopSec + (REAL8)1.0e-9*(REAL8)params->times.stopNan));
               fprintf(stdout,"Difference         = %g \n", (REAL8)((REAL8)params->times.stopSec + (REAL8)1.0e-9*(REAL8)params->times.stopNan)
	                 -  (REAL8)((REAL8)params->times.startSec + params->times.timeStepSize*(REAL8)nData) );
               fflush(stdout);
        #endif

	/* "Incorrect input data stop time" */ /* NOTE 1.0e-10 SAFETY FACTOR */
	/* if ( fabs( (REAL8)((REAL8)params->times.startSec + params->times.timeStepSize*(REAL8)(nData - 1))
		   - (REAL8)((REAL8)params->times.stopSec + (REAL8)1.0e-9*(REAL8)params->times.stopNan) ) > 1.e-10*params->times.timeStepSize ) */
	/* 10/11/01 gam */
	/* 11/17/01 gam */ /* Do not vet stop time since it can be ambigous, e.g, can = start time + N*deltaT or start time + (N - 1)*deltaT. */
	/*
	if ( fabs( (REAL8)((REAL8)params->times.startSec + params->times.timeStepSize*(REAL8)nData)
	   - (REAL8)((REAL8)params->times.stopSec + (REAL8)1.0e-9*(REAL8)params->times.stopNan) )
	   > 1.0e-10*((REAL8)((REAL8)params->times.startSec + params->times.timeStepSize*(REAL8)nData)) ) {
   		ABORT( status, KNOWNPULSARDEMODH_ESTOPTIME, KNOWNPULSARDEMODH_MSGESTOPTIME);
	}
	*/

   } /* END if (params->generateTestData == 'n') */


   #ifdef DEBUG_QUALITYCHANNEL_CODE
     fprintf(stdout, "\nQuality Channel Info:  qc1Found = %i, qc1Index = %i \n",qc1Found,qc1Index);
     fflush(stdout);
   #endif

   /* 11/27/01 gam */ /* Handle the quality channel */ /* IN THE FUTURE ADD gapsFound code here too. */
   if (qc1Found) {

      gpsTimeInterval qc1Times = inout->sequences[qc1Index].range.dTime; /* The GPS start and end time of the quality channel */
      REAL8 qc1DeltaT = (REAL8) qc1Times.timeStepSize;                   /* Quality channel time step size */
      UINT4 gwSRateToqc1SRate = (UINT4)floor(qc1DeltaT/params->deltaT);  /* Ratio of gravity wave sample rate to quality channel sample rate */
      REAL4 gwMeanValue = 0.0;                                           /* Mean value with cutoffs stored in the gw channel */
      UINT4 qc1InputN = inout->sequences[qc1Index].dimensions[0];        /* Number of data points in quality channel from start time */
      UINT4 kpadStart = 0;
      UINT4 kpadEnd = 0;
      /* REAL4 gwChannelMinExpected = (REAL4)GW_CHANNEL_MIN_VALUE_EXPECTED; */ /* 02/07/02 gam */ /* disable use of cutoffs for now */
      /* REAL4 gwChannelMaxExpected = (REAL4)GW_CHANNEL_MAX_VALUE_EXPECTED; */ /* 02/07/02 gam */ /* disable use of cutoffs for now */
      REAL4 gwChannelSum = 0.0;  /* Sum of points of good quality */ /* And not drop out points in the future */
      UINT4 gwChannelCount = 0;  /* Count points of good quality */ /* And not drop out points in the future */

      iStartTime = 0;  /* Reinitialize */
      /* For now assume that quality channel starts at GPS time requested */
      /* GetStartTimeIndex(&qc1Times, &params->gpsStartTime, &iStartTime); */  /* Find the index that corresponds to the starting time */

      /* CalcMeanWithCutoffs( params->vecInputData, (REAL4)GW_CHANNEL_MIN_VALUE_EXPECTED, (REAL4)GW_CHANNEL_MAX_VALUE_EXPECTED, &gwMeanValue); */
      /* Calculate the mean, avoiding points of low quality and using cutoffs */
      /* Outer loop goes throught the quality channel, inner loop sums data for each point of good quality */
      for (i = iStartTime; i < qc1InputN; ++i)
      {
	    #ifdef DEBUG_QUALITYCHANNEL_CODE
              fprintf(stdout, "In sum code: i= %i, qc1 = %g \n",i, inout->sequences[qc1Index].data.real4[i]);
              fflush(stdout);
	    #endif

	    /* Check if quality OK */
      	    if (inout->sequences[qc1Index].data.real4[i] > 0.0) {

	       kpadStart = (i - iStartTime)*gwSRateToqc1SRate;
	       if (kpadStart >= params->vecInputData->length) break; /* We are past the end of the input gw data */
	       kpadEnd = (i + 1 - iStartTime)*gwSRateToqc1SRate;
	       if (kpadEnd > params->vecInputData->length) kpadEnd = params->vecInputData->length;

	      /* Sum up the good quality data */
              for (k = kpadStart; k < kpadEnd; ++k)
              {
                    /* 02/07/02 gam Fix bug in quality channel code; cutoffs were incorrectly applied; disable use of cutoffs for now. */
		    /* fabs was missing from if below.  However, not sure how fabs which for REAL8 casts REAL4 input. */ /* 02/07/02 gam */
    	            /* if ( fabs(params->vecInputData->data[k]) >= gwChannelMinExpected && fabs(params->vecInputData->data[k]) <= gwChannelMaxExpected ) { */
                       #ifdef DEBUG_QUALITYCHANNEL_CODE
                          fprintf(stdout, "k= %i, Summing vecInputData->data[k] = %g \n",k, params->vecInputData->data[k]);
                          fflush(stdout);
                       #endif
                       gwChannelSum += params->vecInputData->data[k];
		       gwChannelCount += 1;
                    /* } */ /* 02/07/02 gam */
              }
	      if (kpadEnd >= params->vecInputData->length) break; /* We are past the end of the input gw data */
	    }

      } /* END for (i = iStartTime; i < qc1InputN; ++i) */

      params->percentPad = (REAL4)( ( ((REAL4)params->vecInputData->length) - ((REAL4)gwChannelCount) )/((REAL4)params->vecInputData->length) );

      if (gwChannelCount > 0) {
      	gwMeanValue = gwChannelSum/((REAL4)gwChannelCount);
      } else {
      	gwMeanValue = 0.0;  /* In the future should abort here ! */
      }

      #ifdef DEBUG_QUALITYCHANNEL_CODE
          fprintf(stdout, "\nqcDeltaT = %g \n",qc1DeltaT);
          fprintf(stdout, "gwSRateToqc1SRate = %i \n",gwSRateToqc1SRate);
          fprintf(stdout, "Quality Channel iStartTime = %i \n",iStartTime);
          fprintf(stdout, "Quality Channel qc1InputN = %i \n",qc1InputN);
          fprintf(stdout, "percentPad = %g \n",params->percentPad);
          fprintf(stdout, "gwMeanValue = %g \n\n",gwMeanValue);
          fflush(stdout);
      #endif

      /* Pad the data of poor quality */
      /* Outer loop goes throught the quality channel, inner loop pads data for each point of poor quality */
      for (i = iStartTime; i < qc1InputN; ++i)
      {
	    #ifdef DEBUG_QUALITYCHANNEL_CODE
              fprintf(stdout, "In pad code: i= %i, qc1 = %g \n",i, inout->sequences[qc1Index].data.real4[i]);
              fflush(stdout);
	    #endif

      	    /* if (inout->sequences[qc1Index].data.real4[i] == 0.0) {} */ /* 02/07/02 gam */
      	    if (inout->sequences[qc1Index].data.real4[i] <= 0.0) {

	       kpadStart = (i - iStartTime)*gwSRateToqc1SRate;
	       if (kpadStart >= params->vecInputData->length) break; /* We are past the end of the input gw data */
	       kpadEnd = (i + 1 - iStartTime)*gwSRateToqc1SRate;
	       if (kpadEnd > params->vecInputData->length) kpadEnd = params->vecInputData->length;

	      /* Pad the low quality data with the mean value */
              for (k = kpadStart; k < kpadEnd; ++k)
              {
                 #ifdef DEBUG_QUALITYCHANNEL_CODE
                      fprintf(stdout, "k= %i, replacing vecInputData->data[k] = %g \n",k, params->vecInputData->data[k]);
                      fflush(stdout);
	         #endif
	         params->vecInputData->data[k] = gwMeanValue;
              }
	      if (kpadEnd >= params->vecInputData->length) break; /* We are past the end of the input gw data */
	    }

      } /* END for (i = iStartTime; i < qc1InputN; ++i) */

   } /* END if (qc1Found) */

 }
 /* END if (params->searchMaster) */

  RETURN( status );
}

/* Apply search routine */
void
LALApplySearch(
    LALStatus             *status,
    LALSearchOutput       *output,
    LALSearchInput        *input,
    LALApplySearchParams  *applyParams
    )
{
  DemodSearchParams* const params = (DemodSearchParams *) applyParams->searchParams;

  UINT4          i = 0;         /* all purpose index */
  UINT4          k = 0;         /* another all purpose index */

  #ifdef DEBUG_KNOWNPULSARDEMODOUTPUTFILES
   FILE           *inputDataDat = NULL;
   FILE           *outputDataDat = NULL;
  #endif

  #ifdef DEBUG_OUTPUT_SFT_FILES
   FILE           *SFT0Dat = NULL;
   FILE           *SFT1Dat = NULL;
   FILE           *SFT2Dat = NULL;
   FILE           *SFT3Dat = NULL;
  #endif

  INITSTATUS( status, "LALApplySearch", KNOWNPULSARDEMODC );
  ATTATCHSTATUSPTR(status);

  if ( !(output) ){
    ABORT( status, KNOWNPULSARDEMODH_ENULL, KNOWNPULSARDEMODH_MSGENULL);
  }

  /* !!!!!!!!!!!!!!!!!!! TO START WITH, WILL RUN CODE ONLY ON THE SEARCH MASTER !!!!!!!!!!!!!!!!! */
  /* !!!!!!! LATER, WILL RUN ON THE SLAVES, AND ADD COMMUNICATION OF RESULTS BACK TO MASTER !!!!! */
  if (params->searchMaster) {

    #ifdef DEBUG_KNOWNPULSARDEMOD
      printf("\nI am the search master.  params->searchMaster = %i \n",params->searchMaster);
      printf("\nRank of this node = %i \n",params->rank);
    #endif

    #ifdef DEBUG_KNOWNPULSARDEMODOUTPUTFILES
  	inputDataDat = fopen("InputData.dat","w");
  	outputDataDat = fopen("OutputData.dat","w");
    #endif

    #ifdef DEBUG_KNOWNPULSARDEMODOUTPUTFILES
       for (k = 0; k < params->vecInputData->length; ++k)
       {
  	   fprintf(stdout, "\nIn ApplySearch: vecInputData[%i] = %g \n",k,params->vecInputData->data[k]);
  	   fflush(stdout);
  	   fprintf(inputDataDat, "%i %g \n",k,fabs(params->vecInputData->data[k]) + 1.0e-6);
  	   fflush(inputDataDat);
       }
    #endif

    /**** START COMPUTING SFTs modified from LALDemodTest.c ****/

	/* 10/15/01 gam */
        /*  Parameters below are for keeping track of which SFTs to do, and if all SFTs are done; used with params->mCohSFTPerCall */
        /*  params->firstK;  */ /* Index of the first SFT to compute and store in the SFTData structure this call to ApplySearch */
        /*  params->lastK;  */  /* Index of the last SFT to compute and store in the SFTData Structure this call to ApplySearch */
        /*  params->finishedSFTs; */ /* Set equal to true when all SFTS for this job have been computed */

        if (!params->finishedSFTs) {
	   params->lastK = params->firstK + params->mCohSFTPerCall;
	   if (params->lastK > params->mObsSFT) params->lastK = params->mObsSFT;
    	   /* for(k=0; k<params->mObsSFT; k++) */
    	   for(k=params->firstK; k<params->lastK; k++)
    	   {

		for(i=0;i<params->nSFT;i++)
		{
			/*
			params->tvec->data[i].re = params->vecInputData->data[i + k*params->nSFT];
			params->tvec->data[i].im = 0.0;
			*/
			params->tvec->data[i] = params->vecInputData->data[i + k*params->nSFT]; /* 09/28/01 gam */

			#ifdef DEBUG_SFT_CODE
			  /* fprintf(stdout,"k = %i, params->tvec->data[%i].re = %g\n",k,i,params->tvec->data[i].re); */
			  fprintf(stdout,"k = %i, tvec->data[%i] = %g\n",k,i,params->tvec->data[i]);/* 09/28/01 gam */
   			  fflush(stdout);
			#endif
		}

		/* Perform FFTW-LAL Fast Fourier Transform */
		/* LALCOMPLEX8VectorFFT(status->statusPtr, params->fvec, params->tvec, params->pfwd); */
    		LALForwardRealFFT( status->statusPtr, params->fvec, params->tvec, params->pfwdReal );

		/* write the data to the SFTData structure and normalize */
		/* for(i=0;i<(int)params->fvec->length;i++) */ /* 11/02/01 gam */ /* i is of time UINT4 */
		for(i=0;i<(UINT4)params->fvec->length;i++)
		{
                        /* 05/14/02 gam */
			/* params->SFTData[k]->fft->data->data[i].re=(REAL8)params->fvec->data[i].re*params->deltaT/sqrt(params->tSFT);
			params->SFTData[k]->fft->data->data[i].im=(REAL8)params->fvec->data[i].im*params->deltaT/sqrt(params->tSFT); */
			params->SFTData[k]->fft->data->data[i].re=(REAL4)(((REAL8)params->fvec->data[i].re)*params->deltaT/sqrt(params->tSFT));
			params->SFTData[k]->fft->data->data[i].im=(REAL4)(((REAL8)params->fvec->data[i].im)*params->deltaT/sqrt(params->tSFT));
		}

 		#ifdef DEBUG_SFT_CODE
		  for(i=0;i<params->fvec->length;i++)
		  {
			fprintf(stdout,"k = %i i = %i f = %g ps = %g\n",k,i,(params->ifMin+i)*params->dfSFT,params->SFTData[k]->fft->data->data[i].re*params->SFTData[k]->fft->data->data[i].re+params->SFTData[k]->fft->data->data[i].im*params->SFTData[k]->fft->data->data[i].im);
			fflush(stdout);
		  }
                #endif

		#ifdef DEBUG_OUTPUT_SFT_FILES
	         switch(k) {
			case 0: SFT0Dat = fopen("SFT0.dat","w"); break;
			case 1: SFT1Dat = fopen("SFT1.dat","w"); break;
			case 2: SFT2Dat = fopen("SFT2.dat","w"); break;
			case 3: SFT3Dat = fopen("SFT3.dat","w"); break;
		 }
		 for(i=0;i<params->fvec->length;i++) {
	            switch(k) {
			case 0: fprintf(SFT0Dat,"%f\t%f\n",params->SFTData[k]->fft->data->data[i].re,params->SFTData[k]->fft->data->data[i].im); fflush(SFT0Dat); break;
			case 1: fprintf(SFT1Dat,"%f\t%f\n",params->SFTData[k]->fft->data->data[i].re,params->SFTData[k]->fft->data->data[i].im); fflush(SFT1Dat); break;
			case 2: fprintf(SFT2Dat,"%f\t%f\n",params->SFTData[k]->fft->data->data[i].re,params->SFTData[k]->fft->data->data[i].im); fflush(SFT2Dat); break;
			case 3: fprintf(SFT3Dat,"%f\t%f\n",params->SFTData[k]->fft->data->data[i].re,params->SFTData[k]->fft->data->data[i].im); fflush(SFT3Dat); break;
		    }
		 }
	         switch(k) {
			case 0: fclose(SFT0Dat); break;
			case 1: fclose(SFT1Dat); break;
			case 2: fclose(SFT2Dat); break;
			case 3: fclose(SFT3Dat); break;
		 }
		#endif

		/* assign particulars to each SFT */
		params->SFTData[k]->fft->epoch=params->timeStamps[k];
		params->SFTData[k]->fft->f0=params->fMin;
		params->SFTData[k]->fft->deltaF=params->dfSFT;
		/* params->SFTData[k]->fft->data->length=params->nSFT; */
		params->SFTData[k]->fft->data->length=params->nDeltaF; /* 09/28/01 gam */
		/* 11/10/01 gam */ /* Calculate stat associated with this SFT */
		CalcDemodPowerStats( params->SFTData[k]->fft->data, params->SFTStats[k], params->fMin, params->dfSFT);
	   } /* END for(k=params->firstK; k<params->lastK; k++) */
        } /* END if (!params->finishedSFTs)  */
	params->firstK = params->lastK;
        if (params->firstK >= params->mObsSFT) params->finishedSFTs = 1;

	/* Finished computing SFTS; perform cleanup tasks */
	/* 10/12/01 gam */
	if (params->finishedSFTs) {
	   /* Set up units to agree with normalization of SFT data above */

           /* 01/28/02 gam */ /* Will need to fix code below to use new definition of LALUnitPair */
           /* 04/16/02 gam */ /* Complete code change to fix code to use new definition of LALUnitPair */

	   LALUnitPair  unitPair;   /* A pair of LALUnits */
	   RAT4         unitPower;  /* Rational number and demoninator minus one to compute power on a LALUnit */
           LALUnit     unitTmp;     /* 04/16/02 gam */ /* Temporary place holder for units. */

  	   /* unitPair.unitOne = lalADCCountUnit; */

	   /* 01/28/02 gam */ /* Will need to fix code below to use new definition of LALUnitPair */
           /* 04/16/02 gam */ /* Complete code change to fix code to use new definition of LALUnitPair */
	   /* unitPair.unitOne = params->unitSFT; */ /* initial unitSFT is set up in ConditionData; Now add in units due to normalization */
	   unitPair.unitOne = &(params->unitSFT);  /* 04/16/02 gam */ /* initial unitSFT is set up in ConditionData; Now add in units due to normalization */
  	   unitPower.numerator = -1;
  	   unitPower.denominatorMinusOne = 1;
  	   /* LALUnitRaise(status->statusPtr, &(unitPair.unitTwo), &lalHertzUnit, &unitPower ); */
  	   LALUnitRaise(status->statusPtr, &unitTmp, &lalHertzUnit, &unitPower ); /* 04/16/02 gam */
    	   CHECKSTATUSPTR (status);
	   unitPair.unitTwo = &unitTmp; /* 04/16/02 gam */
  	   LALUnitMultiply(status->statusPtr, &params->unitSFT, &unitPair);
    	   CHECKSTATUSPTR (status);

	   /* Free memory */
	   LALCDestroyVector(status->statusPtr, &params->fvec);
	   /* LALCDestroyVector(status->statusPtr, &params->tvec); */
	   LALSDestroyVector(status->statusPtr, &params->tvec);  /* 09/28/01 gam */
	   /* LALDestroyComplexFFTPlan(status->statusPtr, &params->pfwd); */
           LALDestroyRealFFTPlan( status->statusPtr, &params->pfwdReal ); /* 09/28/01 gam */
	}

    /**** END COMPUTING SFTs modified from LALDemodTest.c ****/

 #ifdef INCLUDE_DEMOD_CODE
    /**** LALDemod CODE SECTION ENDS HERE ****/
    /* if (params->doDemod) */ 	/* 10/15/01 gam */
    if (params->doDemod && params->finishedSFTs) {
    /**** START COMPUTE demParams based on LALDemodTest.c ****/

        /* variables for computing DeFTs */
	FFT **xHat;                   /* Container for output from LALDemod */
	DemodPar *demParams;          /* Parameters needed for stitching together SFTs to produce coherent demodulated DFT = DeFT */
	ParameterSet *templateParams; /* For holding sky position and spin down parameters */
	CSParams *csParams;           /* Container for ComputeSky */
	INT4 iSkyCoh;                 /* For ComputeSky */
	/* const CHAR *modulation=NULL; */ /* For ComputeSky */ /* 05/14/02 gam */

	/* variables for computing maximum power, frequency for which maximum power occurs, and mean power */
	REAL8 pw = 0.0;
	INT4 ipwMax = 0;
	REAL8 pwSum= 0.0;
	INT4 ipwF0 = 0;  /* 11/05/01 gam */
	REAL8 pwSum2 = 0.0; /* 11/05/01 gam */
	REAL8 tmpF = 0.0; /* 11/05/01 gam */
	REAL8 tmpdF = 0.0; /* 11/05/01 gam */

	/* Allocate space for template parameters */
	templateParams=LALMalloc(sizeof(ParameterSet));
	templateParams->spind=LALMalloc(sizeof(Spindown));
	templateParams->skyP=LALMalloc(2*sizeof(SkyPos));
	templateParams->spind->spParams=LALMalloc(5*sizeof(REAL8));

	templateParams->spind->m = 5;
	templateParams->spind->spParams[0] = params->templateFreqDeriv0;
	templateParams->spind->spParams[1] = params->templateFreqSecondDeriv0;
	templateParams->spind->spParams[2] = params->templateFreqThirdDeriv0;
	templateParams->spind->spParams[3] = params->templateFreqFourthDeriv0;
	templateParams->spind->spParams[4] = params->templateFreqFifthDeriv0;
	templateParams->skyP->alpha = params->templateRA;
	templateParams->skyP->delta = params->templateDec;

	/* Allocate space and set quantity values for demodulation parameters */

	demParams=(DemodPar *)LALMalloc(sizeof(DemodPar));
	demParams->skyConst=(REAL8 *)LALMalloc((2*templateParams->spind->m*(params->mObsSFT+1)+2*params->mObsSFT+3)*sizeof(REAL8));
	demParams->spinDwn=(REAL8 *)LALMalloc(templateParams->spind->m*sizeof(REAL8));
	demParams->if0Max=params->if0Max;
	demParams->if0Min=params->if0Min;
	demParams->mCohSFT=params->mCohSFT;
	demParams->mObsCoh=params->mObsCoh;
	demParams->ifMin=params->ifMin;
	demParams->spinDwnOrder=templateParams->spind->m;

	/* Changed this from signalParams to templateParams */
       	for(i=0;i<templateParams->spind->m;i++)
	{
		demParams->spinDwn[i]=templateParams->spind->spParams[i];
	}

        #ifdef DEBUG_DEMOD_CODE
         for(i=0;i<demParams->spinDwnOrder;i++)
         {
      	   fprintf(stdout,"demParams->spinDwn[%i] = %g \n",i,demParams->spinDwn[i]);
      	   fflush(stdout);
         }
        #endif


	/* Allocate space and set quantities for call to ComputeSky() */
	csParams=(CSParams *)LALMalloc(sizeof(CSParams));
	csParams->skyPos=(REAL8 *)LALMalloc(2*sizeof(REAL8));
	csParams->skyPos[0]=templateParams->skyP->alpha;
	csParams->skyPos[1]=templateParams->skyP->delta;
	csParams->tGPS=params->timeStamps;
	csParams->spinDwnOrder=templateParams->spind->m;
	csParams->mObsSFT=params->mObsSFT;
	csParams->tSFT=params->tSFT;

	/* 05/14/02 gam */ /* No longer used */
	/* csParams->sw=modulation;
	csParams->funcName=params->funcName; */

	iSkyCoh=0;  /* Index for sky patch NOT USED IN THIS CODE */

        for(i=0;i<(2*templateParams->spind->m*(params->mObsSFT+1)+2*params->mObsSFT+3);i++)
        {
      	  demParams->skyConst[i] = 0.0; /* Initialize */
        }


	/* Call COMPUTESKY() */
	ComputeSky(status->statusPtr, demParams->skyConst, iSkyCoh, csParams);
    	CHECKSTATUSPTR (status);

	/* Deallocate space for ComputeSky parameters */
	LALFree(csParams->skyPos);
	LALFree(csParams);

    #ifdef DEBUG_DEMOD_CODE
      for(i=0;i<(2*templateParams->spind->m*(params->mObsSFT+1)+2*params->mObsSFT+3);i++)
      {
      	fprintf(stdout,"demParams->skyConst[%i] = %g \n",i,demParams->skyConst[i]);
      	fflush(stdout);
      }
    #endif

    /**** END COMPUTE demParams based on LALDemodTest.c ****/

    /* Allocate memory for demodulated data */

    xHat=(FFT **)LALMalloc(params->mObsCoh*sizeof(FFT *));
    for(i=0; i<params->mObsCoh; i++)
    {
	   xHat[i]=(FFT *)LALMalloc(sizeof(FFT));
           /* 05/14/02 gam */
	   /*xHat[i]->fft=(COMPLEX16FrequencySeries *)LALMalloc(sizeof(COMPLEX16FrequencySeries));
	   xHat[i]->fft->data=(COMPLEX16Vector *)LALMalloc(sizeof(COMPLEX16Vector));
	   xHat[i]->fft->data->data=(COMPLEX16*)LALMalloc((params->nDeFT)*sizeof(COMPLEX16)); */
	   xHat[i]->fft=(COMPLEX8FrequencySeries *)LALMalloc(sizeof(COMPLEX8FrequencySeries));
	   xHat[i]->fft->data=(COMPLEX8Vector *)LALMalloc(sizeof(COMPLEX8Vector));
	   xHat[i]->fft->data->data=(COMPLEX8*)LALMalloc((params->nDeFT)*sizeof(COMPLEX8));

    }

    k = 0; /* Need to set index of which coherent DFT = DeFT */
    #ifdef DEBUG_DEMOD_CODE
      for(i=0;i<params->nDeFT;i++)
      {
	 xHat[k]->fft->data->data[i].re = 0.0;
	 xHat[k]->fft->data->data[i].im = 0.0;
      }
    #endif
    #ifdef DEBUG_DEMOD_CODE
      for(i=0;i<params->nDeFT;i++)
      {
	 fprintf(stdout,"initialized xHat[%i] f = %g xHat.re = %g xHat.im = %g\n",i,params->f0Min+(REAL8)i/params->tCoh,
		xHat[k]->fft->data->data[i].re, xHat[k]->fft->data->data[i].im);
	 fflush(stdout);
	 if (params->f0Min+(REAL8)i/params->tCoh > params->templateFreq0) break;
      }
    #endif
    demParams->iCoh=k; /* Index of which coherent DFT = DeFT is being generated; used if calling LALDemod more than once. */

    /**** CALL LALDemod !!!! ****/
/* 05/13/02 gam */
/*    LALDemod(status->statusPtr, xHat, params->SFTData, demParams);
    CHECKSTATUSPTR (status); */

    #ifdef DEBUG_DEMOD_CODE
      fprintf(stdout,"\nparams->nDeFT = %i \n",params->nDeFT);
      fflush(stdout);
      /* for(i=0;i<xHat[k]->fft->data->length;i++) */
      /* NOTE: CANNOT USE ABOVE; LENGTH NEVER INITIALIZED !!! */
      for(i=0;i<params->nDeFT;i++)
      {
	 fprintf(stdout,"%i %g %g %g\n",i,params->f0Min+(REAL8)i/params->tCoh,
		xHat[k]->fft->data->data[i].re, xHat[k]->fft->data->data[i].im);
	 fflush(stdout);
	 /* if (params->f0Min+(REAL8)i/params->tCoh > params->templateFreq0) break; */
      }
    #endif

    /* Analyze results */
    /* Calculate the maximum power and the mean power */
    /* params->pwMax = 0.0; */ /* 11/05/01 gam already initialized */
    pwSum = 0.0;
    pwSum2= 0.0; /* 11/05/01 gam */
    tmpdF = params->f0Max - params->f0Min; /* 11/05/01 gam */
    for(i=0;i<params->nDeFT;i++)
    {
                /* 05/14/02 gam */
		/*
		pw=xHat[k]->fft->data->data[i].re*xHat[k]->fft->data->data[i].re+
				xHat[k]->fft->data->data[i].im*xHat[k]->fft->data->data[i].im; */
		pw=(REAL8)(xHat[k]->fft->data->data[i].re*xHat[k]->fft->data->data[i].re+
				xHat[k]->fft->data->data[i].im*xHat[k]->fft->data->data[i].im);


                tmpF = params->f0Min + ((REAL8)i)/params->tCoh; /* 11/05/01 gam */

		#ifdef DEBUG_DEMOD_CODE
  			fprintf(stdout, "freq, pw = %g %g \n",params->f0Min+(REAL8)i/params->tCoh,pw);
  			fflush(stdout);
		#endif
		#ifdef DEBUG_KNOWNPULSARDEMODOUTPUTFILES
  			fprintf(outputDataDat, "%g %g \n",params->f0Min+(REAL8)i/params->tCoh,pw);
  			fflush(outputDataDat);
    		#endif

	        /* pwSum = pwSum + pw; */ /* 11/05/01 gam */
	        /* pwSum2 = pwSum2 + pw*pw; */ /* 11/05/01 gam */

		if(params->pwMax<pw)
		{
			params->pwMax=pw;
			ipwMax=i;
		}

		/* 11/05/01 gam */
		if( fabs(params->templateFreq0 - tmpF) < tmpdF)
		{
			tmpdF = fabs(params->templateFreq0 - tmpF);
			ipwF0=i;
		}
    }
    /* 11/05/01 gam */ /* Compute mean and std dev for all bins except ipwF0 */
    for(i=0;i<params->nDeFT;i++)
    {
                /* 05/14/02 gam */
		/* pw=xHat[k]->fft->data->data[i].re*xHat[k]->fft->data->data[i].re+
				xHat[k]->fft->data->data[i].im*xHat[k]->fft->data->data[i].im; */
		pw=(REAL8)(xHat[k]->fft->data->data[i].re*xHat[k]->fft->data->data[i].re+
				xHat[k]->fft->data->data[i].im*xHat[k]->fft->data->data[i].im);


		if (i != ipwF0)
		{
	           pwSum = pwSum + pw;
	           pwSum2 = pwSum2 + pw*pw;
		}
    }
    /* 11/05/01 gam */
    /* 05/14/02 gam */
    /* params->pwF0 = pw=xHat[k]->fft->data->data[ipwF0].re*xHat[k]->fft->data->data[ipwF0].re +
				xHat[k]->fft->data->data[ipwF0].im*xHat[k]->fft->data->data[ipwF0].im; */
    params->pwF0 = pw = (REAL8)(xHat[k]->fft->data->data[ipwF0].re*xHat[k]->fft->data->data[ipwF0].re +
				xHat[k]->fft->data->data[ipwF0].im*xHat[k]->fft->data->data[ipwF0].im);

    params->freqPWF0 = params->f0Min + ((REAL8)ipwF0)/params->tCoh;
    params->pwMean = pwSum/((REAL8)(params->nDeFT - 1));
    params->pwStdDev = pwSum2/((REAL8)(params->nDeFT - 1)) - params->pwMean*params->pwMean;
    #ifdef DEBUG_BUILDOUTPUT_CODE
    	fprintf(stdout,"\npwStdDev squared = %g \n",params->pwStdDev);
	fflush(stdout);
    #endif
    if ( params->pwStdDev > 0.0 ) {
    	params->pwStdDev = sqrt( params->pwStdDev );
    } else {
        params->pwStdDev = -1.0; /* Should not happen except due to round off error in test data */
    }
    if (params->pwMean > 0.0) {
    	params->pwF0OverpwMean = params->pwF0/params->pwMean;
    } else {
    	params->pwF0OverpwMean = -1.0;  /* Should not happen */
    }
    /* 05/14/02 gam  */
    /* params->phaseForPWF0 = ComputePhase(xHat[k]->fft->data->data[ipwF0].re, xHat[k]->fft->data->data[ipwF0].im); */
    params->phaseForPWF0 = ComputePhase((REAL8)xHat[k]->fft->data->data[ipwF0].re, (REAL8)xHat[k]->fft->data->data[ipwF0].im);

    #ifdef DEBUG_BUILDOUTPUT_CODE
    	fprintf(stdout,"\nfreqPWF0 = %g, pwF0 = %g, pwMean = %g, pwF0/pwMean = %g \n",params->freqPWF0,params->pwF0,params->pwMean,params->pwF0OverpwMean);
    	fprintf(stdout,"\npwStdDev = %g, phaseForPWF0 = %g \n",params->pwStdDev, params->phaseForPWF0);
	fflush(stdout);
    #endif

    #ifdef DEBUG_KNOWNPULSARDEMODOUTPUTFILES
  	fclose(inputDataDat);
  	fclose(outputDataDat);
    #endif

	/* Deallocate memory used */

	for(i=0; i<params->mObsCoh; i++)
	{
		LALFree(xHat[i]->fft->data->data);
		LALFree(xHat[i]->fft->data);
		LALFree(xHat[i]->fft);
		LALFree(xHat[i]);
	}
	LALFree(xHat);

	/* Deallocate template params */
	LALFree(templateParams->spind->spParams);
	LALFree(templateParams->spind);
	LALFree(templateParams->skyP);
	LALFree(templateParams);

	/* Deallocate demodulation params */
	LALFree(demParams->skyConst);
	LALFree(demParams->spinDwn);
	LALFree(demParams);
 } /* END if (params->doDemod) */
   /**** LALDemod CODE SECTION ENDS HERE ****/
 #endif

   /***** Set output data structure STARTS HERE *****/

    	output->numOutput 	= 0;

	/* 10/15/01 gam */
	if (params->finishedSFTs) {

	  if (params->doDemod) {
    		BuildOutput(status->statusPtr, params, output);
	  } else {
		/* Just output SFTs into frame files (ilwd files for stand alone wrapper */
    		/* BuildFrameOutput(status->statusPtr, params, params->SFTData, params->timeIntervals, output); */
    		BuildFrameOutput(status->statusPtr, params, params->SFTData, params->SFTStats, params->timeIntervals, output);
	  }
    	  CHECKSTATUSPTR (status);

	  LALFree(params->timeStamps);

	  LALFree(params->timeIntervals);

	  /* Deallocate SFTData structure */
	  for(i=0;i<params->mObsSFT;i++)
	  {
		LALFree(params->SFTData[i]->fft->data->data);
		LALFree(params->SFTData[i]->fft->data);
		LALFree(params->SFTData[i]->fft);
		LALFree(params->SFTData[i]);
	  }
	  LALFree(params->SFTData);

          /* 11/10/01 gam */ /* DAllocate memory for the SFTStats structure */
    	  for(i=0;i<params->mObsSFT;i++)
    	  {
		LALFree(params->SFTStats[i]);
    	  } /* END for(i=0;i<params->mObsSFT;i++) */
	  LALFree(params->SFTStats);

	  output->fracRemaining = 0;
    	  output->notFinished	= 0;

	  #ifdef DEBUG_FRACREMAINING_CODE
  	         fprintf(stdout, "\nfracRemaining = %g \n", output->fracRemaining);
  		 fflush(stdout);
	  #endif

	} else {

	      output->fracRemaining = (REAL4)params->lastK/(REAL4)params->mObsSFT;
    	      output->notFinished = 1;

	      #ifdef DEBUG_FRACREMAINING_CODE
  		fprintf(stdout, "\nfracRemaining = %g \n", output->fracRemaining);
  		fflush(stdout);
	      #endif

	} /* END if (params->finishedSFTs) else ... */

   /***** Set output data structure ENDS HERE *****/

  } else {
  /* if NOT (params->searchMaster) */
  /* !!!!!!!!!!!!!!!!!!! TO START WITH, WILL RUN CODE ONLY ON THE SEARCH MASTER !!!!!!!!!!!!!!!!! */
  /* !!!!!!! LATER, WILL RUN ON THE SLAVES, AND ADD COMMUNICATION OF RESULTS BACK TO MASTER !!!!! */

    /* Just continue for now */
    /* Set output data structure */
    output->numOutput 	= 0;
    output->result	= NULL;
    output->fracRemaining = 0;
    output->notFinished	= 0;

  } /* END if (params->searchMaster) else ... */

  CHECKSTATUSPTR (status);
  DETATCHSTATUSPTR (status);
  RETURN( status );
}

/* Finalize search routine */
void
LALFinalizeSearch(
    LALStatus             *status,
    void                 **searchParams
    )
{
  DemodSearchParams* const params = (DemodSearchParams *) *searchParams;

  INITSTATUS( status, "LALFinalizeSearch", KNOWNPULSARDEMODC );
  ATTATCHSTATUSPTR (status);

  /* !!!!!!!!!!!!!!!!!!! TO START WITH, WILL RUN CODE ONLY ON THE SEARCH MASTER !!!!!!!!!!!!!!!!! */
  /* !!!!!!! LATER, WILL RUN ON THE SLAVES, AND ADD COMMUNICATION OF RESULTS BACK TO MASTER !!!!! */
  if (params->searchMaster) {

   LALSDestroyVector(status->statusPtr, &(params->vecInputData));


  }

  /* 10/16/01 gam */ /* 11/05/01 gam */
  LALFree(params->dataConDesc);
  LALFree(params->IFO);
  LALFree(params->targetName);

  LALFree(params->dsoName);

  #ifdef DEBUG_KNOWNPULSARDEMOD
    fprintf(stdout, "\nLALFinalizeSearch, my rank is %i.\n",params->rank);
    fflush(stdout);
  #endif

  LALFree(*searchParams);

  DETATCHSTATUSPTR (status);
  RETURN(status);

}

/********************** local static function definitions ********************/

/* !!!!!!!!!!!!!!! COPIED FROM LALDemodTest !!!!!!!!!!!!! */
/* internal routine to convert detector time to barycentric time */
/* Note: this routine will be replaced by Barycenter(), written  */
/* by C. Cutler, Albert-Einstein-Institut.  			 */
/*  10/08/01 gam; updated latest version form LALDemodTest */
void tdb(REAL8 alpha, REAL8 delta, REAL8 t_GPS, REAL8 *T, REAL8 *Tdot)
{

	/*RA alpha and DEC delta have their usual meanings, but units
  are radians, not hours or degrees */

	REAL8 tdiff;
	REAL8 tdiff_day;
	REAL8 tdiff_yr;
	REAL8 beta;
	REAL8 x;
	REAL8 y;
	REAL8 lam;
	REAL8 phi0;
	REAL8 psi0;
	REAL8 phi;
	REAL8 psi;
	REAL8 cose;
	REAL8 sine;
	REAL8 sind;
	REAL8 cosd;
	REAL8 sina;
	REAL8 cosa;
	REAL8 cosb;

	/*phi0 and psi0 given location and orientation, resp., of Earth at t_GPS =0 */
	REAL8 T_bary, Tdot_bary;
	/* T_bary (what we want to know) is pulse emission time in TDB coords.*/
	/* The lab measures GPS time, but we set t_GPS =0 at first instant of 1998*/
	REAL8 AU = 1.49597870e13/2.99792458e10;
	REAL8 RE = 6.37814e8/2.99792458e10;
	REAL8 eps = (23.0 + 26.0/60 + 21.488/3600)*LAL_PI/180;
	/*    printf("eps is %lf \n",eps); */
	phi0=2.86e0;
	psi0=4.12e0;
	/* Those values of phi0 and psi0 are completely made up */

	/* Now converting from equatorial to ecliptic coords. */
	sine=sin(eps);
	cose=cos(eps);
	sind=sin(delta);
	cosd=cos(delta);
	sina=sin(alpha);
	cosa=cos(alpha);
	beta=cose*sind-sine*cosd*sina;
	cosb=sqrt(1.0-beta*beta);
	x=cosd*cosa/cosb;
	y=(sine*sind+cose*cosd*sina)/cosb;
	lam=atan2(y,x);
	/*  printf("beta, lam are (rad) %lf  %lf \n", beta, lam);  */
	phi=phi0+2*LAL_PI*t_GPS/3.15576e7;
	psi=psi0+2*LAL_PI*t_GPS/8.64e4;
	tdiff_day=RE*cosd*cos(psi-alpha);
	tdiff_yr=AU*cosb*cos(phi-lam);
	tdiff=tdiff_day+tdiff_yr;
	T_bary=t_GPS+tdiff;
	Tdot_bary=1.e0-RE*cosd*sin(psi-alpha)*2*LAL_PI/8.64e4
		-AU*cosb*sin(phi-lam)*2*LAL_PI/3.15576e7;

	/* switch to turn mod off/on */
	/* note that 'off' negates the function of the code! */

	*T=T_bary;
 	*Tdot=Tdot_bary;

}

/*
void tdb(REAL8 alpha, REAL8 delta, REAL8 t_GPS, REAL8 *T, REAL8 *Tdot, const CHAR *sw)
{
*/
 /* RA alpha and DEC delta have their usual meanings, but units are radians, not hours or degrees */
/*
	REAL8 tdiff;
	REAL8 tdiff_day;
	REAL8 tdiff_yr;
	REAL8 beta;
	REAL8 x;
	REAL8 y;
	REAL8 lam;
	REAL8 phi0;
	REAL8 psi0;
	REAL8 phi;
	REAL8 psi;
	REAL8 cose;
	REAL8 sine;
	REAL8 sind;
	REAL8 cosd;
	REAL8 sina;
	REAL8 cosa;
	REAL8 cosb;
*/
	/*phi0 and psi0 given location and orientation, resp., of Earth at t_GPS =0 */
/*
	REAL8 T_bary, Tdot_bary;
*/
	/* T_bary (what we want to know) is pulse emission time in TDB coords.*/
	/* The lab measures GPS time, but we set t_GPS =0 at first instant of 1998*/
/*
	REAL8 AU = 1.49597870e13/2.99792458e10;
	REAL8 RE = 6.37814e8/2.99792458e10;
	REAL8 eps = (23.0 + 26.0/60 + 21.488/3600)*LAL_PI/180;
*/
	/*    printf("eps is %lf \n",eps); */
/*
	phi0=2.86e0;
	psi0=4.12e0;
*/
	/* Those values of phi0 and psi0 are completely made up */

	/* Now converting from equatorial to ecliptic coords. */
/*
	sine=sin(eps);
	cose=cos(eps);
	sind=sin(delta);
	cosd=cos(delta);
	sina=sin(alpha);
	cosa=cos(alpha);
	beta=cose*sind-sine*cosd*sina;
	cosb=sqrt(1.0-beta*beta);
	x=cosd*cosa/cosb;
	y=(sine*sind+cose*cosd*sina)/cosb;
	lam=atan2(y,x);
*/
	/*  printf("beta, lam are (rad) %lf  %lf \n", beta, lam);  */
/*
	phi=phi0+2*LAL_PI*t_GPS/3.15576e7;
	psi=psi0+2*LAL_PI*t_GPS/8.64e4;
	tdiff_day=RE*cosd*cos(psi-alpha);
	tdiff_yr=AU*cosb*cos(phi-lam);
	tdiff=tdiff_day+tdiff_yr;
	T_bary=t_GPS+tdiff;
	Tdot_bary=1.e0-RE*cosd*sin(psi-alpha)*2*LAL_PI/8.64e4
		-AU*cosb*sin(phi-lam)*2*LAL_PI/3.15576e7;
*/
	/* switch to turn mod off/on */
	/* note that 'off' negates the function of the code! */
/*
	if(sw!=NULL){
	*T=T_bary;
 	*Tdot=Tdot_bary;}

	else {
	*T=t_GPS;
	*Tdot=1.0;}

	}
*/

/*
void times(REAL8 tSFT, int howMany, LIGOTimeGPS *ts, INT4 sw)
{
	int i=0, j=0;
	int temp1, temp2;
	#ifdef DEBUG_DEMOD_CODE
	  FILE *fp;
	  fp=fopen("times.blah","w");
	#endif
	while(i<sw*howMany)
	{
		temp1=floor(tSFT*i);
		temp2=(int)((tSFT*(double)i-temp1)*1E9);

		if(i==0) {

			ts[j].gpsSeconds=0;
			ts[j].gpsNanoSeconds=0;

		} else {

			ts[j].gpsSeconds=temp1;
			ts[j].gpsNanoSeconds=temp2;
		}
		#ifdef DEBUG_DEMOD_CODE
		  fprintf(fp,"%d %d %d %d %d %d\n",ts[j].gpsSeconds, ts[j].gpsNanoSeconds,
		     timeIntervals[j].startSec, timeIntervals[j].startNan, timeIntervals[j].stopSec,timeIntervals[j].stopNan );
		#endif
		i=i+sw;
		j++;
	}
	#ifdef DEBUG_DEMOD_CODE
	  fclose(fp);
	#endif
}
*/

/* void MoreTimes(REAL8 tSFT, int howMany, LIGOTimeGPS *ts, gpsTimeInterval *timeIntervals, INT4 sw) */ /* 10/10/01 gam */
/* void MoreTimes(REAL8 tSFT, int howMany, UINT4 gpsStartTime, LIGOTimeGPS *ts, gpsTimeInterval *timeIntervals, INT4 sw) */  /* 11/02/01 gam */
void MoreTimes(REAL8 tSFT, INT4 howMany, UINT4 gpsStartTime, LIGOTimeGPS *ts, gpsTimeInterval *timeIntervals, INT4 sw)
{
        /*
	int i=0, j=0;
	int temp1, temp2;
	int temp3, temp4;
	*/

	/* 11/02/01 gam */
        INT4 i=0, j=0;
	INT4 temp1, temp2;
	INT4 temp3, temp4;

	#ifdef DEBUG_DEMOD_CODE
	  FILE *fp;
	  fp=fopen("times.blah","w");
	#endif
	while(i<sw*howMany)
	{
		temp1=floor(tSFT*i);
		/* temp2=(int)((tSFT*(double)i-temp1)*1E9); */
		/* temp2=(INT4)((tSFT*(double)i-temp1)*1E9); */ /* 11/02/01 gam */
		temp2=(INT4)((tSFT*(REAL8)i-temp1)*1E9); /* 11/17/01 gam */

		temp3=floor(tSFT*(i + 1));
		/* temp4=(int)((tSFT*(double)(i + 1) - temp3)*1E9); */
		/* temp4=(INT4)((tSFT*(double)(i + 1) - temp3)*1E9); */ /* 11/02/01 gam */
		temp4=(INT4)((tSFT*(REAL8)(i + 1) - temp3)*1E9); /* 11/17/01 gam */

		if(i==0) {

		        /* 10/10/01 gam */ /*
			ts[j].gpsSeconds=0;
			ts[j].gpsNanoSeconds=0;

			timeIntervals[j].startSec = 0;
			timeIntervals[j].startNan = 0;
			timeIntervals[j].stopSec = temp3;
			timeIntervals[j].stopNan = temp4;
			*/

			ts[j].gpsSeconds = gpsStartTime;
			ts[j].gpsNanoSeconds = 0;

			timeIntervals[j].startSec = gpsStartTime;
			timeIntervals[j].startNan = 0;
			timeIntervals[j].stopSec = gpsStartTime + temp3;
			timeIntervals[j].stopNan = temp4;

		} else {

		        /* 10/10/01 gam */ /*
			ts[j].gpsSeconds=temp1;
			ts[j].gpsNanoSeconds=temp2;

			timeIntervals[j].startSec = temp1;
			timeIntervals[j].startNan = temp2;
			timeIntervals[j].stopSec = temp3;
			timeIntervals[j].stopNan = temp4;
			*/

			ts[j].gpsSeconds = gpsStartTime + temp1;
			ts[j].gpsNanoSeconds = temp2;

			timeIntervals[j].startSec = gpsStartTime + temp1;
			timeIntervals[j].startNan = temp2;
			timeIntervals[j].stopSec = gpsStartTime + temp3;
			timeIntervals[j].stopNan = temp4;

		}
		#ifdef DEBUG_DEMOD_CODE
		  fprintf(fp,"%d %d %d %d %d %d\n",ts[j].gpsSeconds, ts[j].gpsNanoSeconds,
		     timeIntervals[j].startSec, timeIntervals[j].startNan, timeIntervals[j].stopSec,timeIntervals[j].stopNan );
		#endif
		i=i+sw;
		j++;
	}
	#ifdef DEBUG_DEMOD_CODE
	  fclose(fp);
	#endif
}

/* For building output for the database */
/* Copied and modified from Moments.c   */
/* Author: Charlton, Philip; Brady, P R */
/*
 * Function BuildOutput
 *
 * Create a linked list containing the database entries and put it in
 * the search output structure
 */
static
void
BuildOutput(LALStatus*            const status,
	    const DemodSearchParams* const params,
	    LALSearchOutput*      const output)
{

  /* const CHAR* const PROGRAM = "KnownPulsarDemod"; */ /* 10/16/01 gam */
  UINT4  i = 0;        /* all purpose index */

  dataBase* dbEntry = 0;
  outPut* outputArray = 0;

  INITSTATUS( status, "BuildOutput", KNOWNPULSARDEMODC );

  /* Number of output objects */
  output->numOutput = 1;  /* FOR NOW HARD CODE AS 1 */

  #ifdef DEBUG_BUILDOUTPUT_CODE
    	fprintf(stdout,"\nIn BuildOutput. output->numOutput = %i \n",output->numOutput);
	fflush(stdout);
  #endif

  /* make check fails if below is used */
  /* outputArray = output->result = (outPut *) LALMalloc(1 * sizeof(outPut) ); */

  outputArray = output->result = (outPut *) LALCalloc( 1, sizeof(outPut) );
  if ( ! outputArray )
  {
    ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
  }

  /*
   * Start a linked list of database entries - the head of the list
   * starts with the "results" field
   */

  for (i = 0; i < output->numOutput; i++) {

      /*  outputArray[0].templateNumber = 0; */    /* not relevant */
      outputArray[0].search         = periodic;
      outputArray[0].significant    = 1;   /* 1 - IS significant */
      outputArray[0].states         = 0;
      outputArray[0].optional       = 0;

      /* Just set equal to zero for now */
      outputArray[i].indexNumber = 0;

      /* Put entry into the SNGL_DPERIODIC database table */
      /* DPERIODIC = DIRECTED PERIODIC */

      /* Initiate linked list */
      dbEntry = outputArray[i].results = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->previous = NULL;

      /* creator_db */
      /* 10/16/01 gam */
      /*
      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "creator_db" );

      dbEntry->type       = int_4s;
      dbEntry->numberRows = 1;
      dbEntry->rows.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! dbEntry->rows.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.int4s) = 1;
      */

      /* process_id */
      /* 10/16/01 gam */
      /*
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "process_id" );

      dbEntry->type       = char_s_ptr;
      dbEntry->numberRows = 1;
      dbEntry->rows.chars = (CHAR *) LALMalloc( (strlen( PROGRAM ) + 1) * sizeof(CHAR) );
      if ( ! dbEntry->rows.chars )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      strcpy(dbEntry->rows.chars, PROGRAM);
      dbEntry->rowDimensions = (UINT4 *) LALMalloc( sizeof(UINT4) );
      if ( ! dbEntry->rowDimensions )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      */
      /* Don't include terminating null */
      /*
      *(dbEntry->rowDimensions) = (UINT4) strlen( PROGRAM );
      */

      /* filter_id */
      /* 10/16/01 gam */ /* THIS IS NOW THE FIRST ENTRY; dbEntry initialized above */
      /*
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;
      */

      /*  10/18/01 gam */
      /*
      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "filter_id" );

      dbEntry->type       = char_s_ptr;
      dbEntry->numberRows = 1;
      */
      /* dbEntry->rows.chars = (CHAR *) LALMalloc( (strlen( "FILTER_ID" ) + 1) * sizeof(CHAR) ); */
      /* dbEntry->rows.chars = (CHAR *) LALMalloc( (strlen(params->filterID) + 1) * sizeof(CHAR) ); */ /* 10/16/01 gam */
      /*
      if ( ! dbEntry->rows.chars )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      */
      /* strcpy(dbEntry->rows.chars, "FILTER_ID"); */
      /* strcpy(dbEntry->rows.chars, params->filterID); */ /* 10/16/01 gam */
      /*
      dbEntry->rowDimensions = (UINT4 *) LALMalloc( sizeof(UINT4) );
      if ( ! dbEntry->rowDimensions )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      */
      /* Don't include terminating null */
      /* *(dbEntry->rowDimensions) = (UINT4) strlen( "FILTER_ID" ); */
      /* *(dbEntry->rowDimensions) = (UINT4) strlen( params->filterID ); */

      /* ifo */  /*  10/18/01 gam */ /* dbEntry is initialized above */
      /*
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;
      */

      /* ifo */
      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "ifo" );

      dbEntry->type       = char_s_ptr;
      dbEntry->numberRows = 1;
      /* dbEntry->rows.chars = (CHAR *) LALMalloc( (strlen( "??" ) + 1) * sizeof(CHAR) ); */
      dbEntry->rows.chars = (CHAR *) LALMalloc( (strlen( params->IFO ) + 1) * sizeof(CHAR) ); /* 10/16/01 gam */
      if ( ! dbEntry->rows.chars )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      /* strcpy(dbEntry->rows.chars, "??"); */
      strcpy(dbEntry->rows.chars, params->IFO); /* 10/16/01 gam */
      dbEntry->rowDimensions = (UINT4 *) LALMalloc( sizeof(UINT4) );
      if ( ! dbEntry->rowDimensions )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      /* Don't include terminating null */
      /* *(dbEntry->rowDimensions) = (UINT4) strlen( "??" ); */
      *(dbEntry->rowDimensions) = (UINT4) strlen( params->IFO ); /* 10/16/01 gam */


      /* search */ /* 12/13/01 gam */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "search" );

      dbEntry->type       = char_s_ptr;
      dbEntry->numberRows = 1;
      dbEntry->rows.chars = (CHAR *) LALMalloc( (strlen( "knownpulsardemod" ) + 1) * sizeof(CHAR) );
      if ( ! dbEntry->rows.chars )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      strcpy(dbEntry->rows.chars, "knownpulsardemod");
      dbEntry->rowDimensions = (UINT4 *) LALMalloc( sizeof(UINT4) );
      if ( ! dbEntry->rowDimensions )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      /* Don't include terminating null */
      *(dbEntry->rowDimensions) = (UINT4) strlen( "knownpulsardemod" );

      /* start_time */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "start_time" );

      dbEntry->type       = int_4s;
      dbEntry->numberRows = 1;
      dbEntry->rows.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! dbEntry->rows.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.int4s) = params->times.startSec;

     /* start_time_ns */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "start_time_ns" );

      dbEntry->type       = int_4s;
      dbEntry->numberRows = 1;
      dbEntry->rows.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! dbEntry->rows.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.int4s) = params->times.startNan;

     /* end_time */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "end_time" );

      dbEntry->type       = int_4s;
      dbEntry->numberRows = 1;
      dbEntry->rows.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! dbEntry->rows.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.int4s) = params->times.stopSec;

     /* end_time_ns */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "end_time_ns" );

      dbEntry->type       = int_4s;
      dbEntry->numberRows = 1;
      dbEntry->rows.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! dbEntry->rows.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.int4s) = params->times.stopNan;

      /* duration */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "duration" );

      /* 10/17/01 gam */
      dbEntry->type       = real_4;
      dbEntry->numberRows = 1;
      dbEntry->rows.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! dbEntry->rows.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      /* *(dbEntry->rows.real4) = params->deltaT*params->inputN; */
      /* *(dbEntry->rows.real4) = params->deltaT*((REAL4)(params->inputN - 1)); */
      *(dbEntry->rows.real4) = (REAL4)(((REAL4)params->deltaT)*((REAL4)params->inputN)); /* 10/17/01 gam */

      /* 10/10/01 gam */
      /* 10/17/01 gam */
      /*
      dbEntry->type       = real_8;
      dbEntry->numberRows = 1;
      dbEntry->rows.real8 = (REAL8 *) LALMalloc( sizeof(REAL8) );
      if ( ! dbEntry->rows.real8 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      */
      /* *(dbEntry->rows.real8) = params->deltaT*((REAL8)(params->inputN - 1)); */
      /* *(dbEntry->rows.real8) = params->deltaT*((REAL8)params->inputN); */ /* 10/11/01 gam */

      /* target_name */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "target_name" );

      dbEntry->type       = char_s_ptr;
      dbEntry->numberRows = 1;
      /* dbEntry->rows.chars = (CHAR *) LALMalloc( (strlen( "A known pulsar" ) + 1) * sizeof(CHAR) ); */
      dbEntry->rows.chars = (CHAR *) LALMalloc( (strlen( params->targetName ) + 1) * sizeof(CHAR) ); /* 10/16/01 gam */
      if ( ! dbEntry->rows.chars )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      /* strcpy(dbEntry->rows.chars, "A known pulsar"); */
      strcpy(dbEntry->rows.chars, params->targetName); /* 10/16/01 gam */
      dbEntry->rowDimensions = (UINT4 *) LALMalloc( sizeof(UINT4) );
      if ( ! dbEntry->rowDimensions )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      /* Don't include terminating null */
      /* *(dbEntry->rowDimensions) = (UINT4) strlen( "A known pulsar" ); */
      *(dbEntry->rowDimensions) = (UINT4) strlen( params->targetName ); /* 10/16/01 gam */

      /* sky_ra */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "sky_ra" );

      dbEntry->type       = real_8;
      dbEntry->numberRows = 1;
      dbEntry->rows.real8 = (REAL8 *) LALMalloc( sizeof(REAL8) );
      if ( ! dbEntry->rows.real8 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.real8) = params->templateRA;

      /* sky_dec */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "sky_dec" );

      dbEntry->type       = real_8;
      dbEntry->numberRows = 1;
      dbEntry->rows.real8 = (REAL8 *) LALMalloc( sizeof(REAL8) );
      if ( ! dbEntry->rows.real8 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.real8) = params->templateDec;

      /* frequency */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "frequency" );

      dbEntry->type       = real_8;
      dbEntry->numberRows = 1;
      dbEntry->rows.real8 = (REAL8 *) LALMalloc( sizeof(REAL8) );
      if ( ! dbEntry->rows.real8 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.real8) = (REAL8) params->freqPWF0; /* 11/05/01 gam */

      /* amplitude */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "amplitude" );

      dbEntry->type       = real_4;
      dbEntry->numberRows = 1;
      dbEntry->rows.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! dbEntry->rows.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.real4) = (REAL4) sqrt(params->pwF0);  /* 11/05/01 gam */

      /* phase */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "phase" );

      dbEntry->type       = real_4;
      dbEntry->numberRows = 1;
      dbEntry->rows.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! dbEntry->rows.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      *(dbEntry->rows.real4) = (REAL4) params->phaseForPWF0;  /* 11/05/01 gam */

      /* snr */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "snr" );

      dbEntry->type       = real_4;
      dbEntry->numberRows = 1;
      dbEntry->rows.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! dbEntry->rows.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      /* *(dbEntry->rows.real4) = (REAL4) params->pwF0OverpwMean; */ /* 11/05/01 gam */
      if (params->pwStdDev > 0.0) {
       	 *(dbEntry->rows.real4) = (REAL4)((params->pwF0 - params->pwMean)/params->pwStdDev);
      } else {
       	 *(dbEntry->rows.real4) = -1.0;   /* This case should not happen except in test data; hard code for now */
      }

      /* confidence */
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "confidence" );

      dbEntry->type       = real_4;
      dbEntry->numberRows = 1;
      dbEntry->rows.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! dbEntry->rows.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->rowDimensions = NULL;
      /* *(dbEntry->rows.real4) = 0.0; */
      *(dbEntry->rows.real4) = 1.0 - exp(-1.0*params->pwF0OverpwMean); /* 11/05/01 gam */

      /* event_id */
      /* FOR NOW HARD CODE AS EVENT_ID */
      /* 10/16/01 gam */
      /*
      dbEntry->next = (dataBase *) LALMalloc( sizeof(dataBase) );
      if ( ! dbEntry->next )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      dbEntry->next->previous = dbEntry;
      dbEntry = dbEntry->next;

      sprintf( dbEntry->tableName, "sngl_dperiodic" );
      sprintf( dbEntry->columnName, "event_id" );

      dbEntry->type       = char_s_ptr;
      dbEntry->numberRows = 1;
      dbEntry->rows.chars = (CHAR *) LALMalloc( (strlen( "EVENT_ID" ) + 1) * sizeof(CHAR) );
      if ( ! dbEntry->rows.chars )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      strcpy(dbEntry->rows.chars, "EVENT_ID");
      dbEntry->rowDimensions = (UINT4 *) LALMalloc( sizeof(UINT4) );
      if ( ! dbEntry->rowDimensions )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      */
      /* Don't include terminating null */
      /* *(dbEntry->rowDimensions) = (UINT4) strlen( "EVENT_ID" ); */

      /* Terminate linked list */
      dbEntry->next = NULL;

  } /* END for (i = 0; i < output->numOutput; i++) */

  RETURN(status);
}

/* Compute the phase of a complex Number */
REAL8 ComputePhase(REAL8 x, REAL8 y)
{
	REAL8 phi = 0.0;
	REAL8 pi = LAL_PI;
	REAL8 twoPi = LAL_TWOPI;
	REAL8 piOverTwo = LAL_PI_2;
        if (x != 0.0) {
		phi = atan(fabs(y/x));
		if (x < 0) {
			if (y < 0) {
				// quad III
				phi = phi + pi;
			} else {
				// quad II
				phi = pi -  phi;
			}
		} else {
			if (y < 0) {
				// quad IV
				phi = twoPi -  phi;
			} else {
				// quad I; just continue.
			}
		}
	} else {
		if (y >= 0.0) {
			phi = piOverTwo;     //return pi/2.
		} else {
			phi = 3.0*piOverTwo; //return 3pi/2.
		}
	}
	return phi;
}

/* Copied and modified from StochasticBuildOutput.c written by: */
/* Anderson, W.G., Romano, J.D., and Whelan, J.T. */
/*void StochasticBuildOutput (
      LALStatus*                  status,
      LALSearchOutput*            output,
      COMPLEX8FrequencySeries*    spec,
      gpsTimeInterval             timeInterval
)
*/
/*
static void
BuildFrameOutput(LALStatus*            const status,
	    const DemodSearchParams*   const params,
	    FFT                        **SFTData,
            gpsTimeInterval            *timeIntervals,
	    LALSearchOutput*           const output)
*/
/* 11/10/01 gam */ /* Add in SFTStats */
static void BuildFrameOutput(LALStatus* const status,
	    const DemodSearchParams*    const params,
	    FFT                         **SFTData,
            DemodPowerStats             **SFTStats,
            gpsTimeInterval             *timeIntervals,
	    LALSearchOutput*            const output)

{
   outPut* outputArray = 0;
   multiDimData*   mddEntry;
   CHARVector     *unitString;

   dcHistory *  historyEntry; /* 11/05/01 gam */

   UINT4 k = 0;    /* index used with SFTData */
   /* UINT4 icopy = 0; */ /* index used to copy SFT data to output structure */ /* 03/06/02 gam */ /* 05/14/02 gam */

   INITSTATUS( status, "BuildFrameOutput", KNOWNPULSARDEMODC );
   ATTATCHSTATUSPTR (status);

    /* output->numOutput = 1; */
    /* output->numOutput = params->inputM; */
    output->numOutput = params->mObsSFT; /* 10/12/01 gam */

   /*
    *
    * create memory for the outPut array
    *
    */

  outputArray = output->result = (outPut *) LALCalloc( output->numOutput, sizeof(outPut) );
  if ( ! outputArray )
  {
    ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
  }

   /*
    *
    * fill the array of output
    *
    */

  for (k = 0; k < output->numOutput; k++) {

      /*  outputArray[0].templateNumber = 0; */    /* not relevant */
      outputArray[k].search         = periodic;
      outputArray[k].significant    = 1;   /* 1 - IS significant */
      outputArray[k].states         = 0;
      outputArray[k].optional       = 0;

      /* Just set equal to zero for now */
      outputArray[k].indexNumber = 0;


     /*
      *
      * alloc memory for the optional linked list
      *
      */

     /* allocate memory */
     /* mddEntry = outputArray->optional = (multiDimData *) LALMalloc (sizeof (multiDimData)); */
     /* Memory allocation fails if above line used in place of line below */
     /* mddEntry = outputArray->optional = (multiDimData *) LALCalloc (1, sizeof(multiDimData)); */
     mddEntry = outputArray[k].optional = (multiDimData *) LALCalloc (1, sizeof(multiDimData));
     if (!mddEntry)
     {
      ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
     }

     /* write the name and domain of the data */
     /* strncpy (mddEntry->name, "filter_output", maxMultiDimName); */
     strncpy (mddEntry->name, "sft_output", maxMultiDimName);

     /* Set the dimensions from the output frequency series */
     unitString = NULL;
     TRY( LALCHARCreateVector(status->statusPtr, &unitString, LALUnitTextSize), status );
     TRY( LALUnitAsString(status->statusPtr, unitString, &(params->unitSFT)), status );
     strncpy (mddEntry->units, unitString->data, maxMultiDimUnits);
     TRY( LALCHARDestroyVector(status->statusPtr, &unitString), status );
     mddEntry->space = freqD;

     #ifdef DEBUG_BUILDFRAMEOUTPUT_CODE
  	fprintf(stdout, "\ngpsStartTimeSec= %i \n", (UINT4)timeIntervals[k].startSec);
  	fprintf(stdout, "\ngpsStartTimeNan = %i \n",(UINT4)timeIntervals[k].startNan);
  	fprintf(stdout, "\ngpsStopTimeSec = %i \n", (UINT4)timeIntervals[k].stopSec);
  	fprintf(stdout, "\ngpsStopTimeNan = %i \n", (UINT4)timeIntervals[k].stopNan);
  	fflush(stdout);
     #endif

     mddEntry->range.dFreq.gpsStartTimeSec = (UINT4)timeIntervals[k].startSec;
     mddEntry->range.dFreq.gpsStartTimeNan = (UINT4)timeIntervals[k].startNan;
     mddEntry->range.dFreq.gpsStopTimeSec = (UINT4)timeIntervals[k].stopSec;
     mddEntry->range.dFreq.gpsStopTimeNan = (UINT4)timeIntervals[k].stopNan;

     #ifdef DEBUG_BUILDFRAMEOUTPUT_CODE
  	fprintf(stdout, "\nstartFreq = %f \n", SFTData[k]->fft->f0);
  	fprintf(stdout, "\nstopFreq = %f \n", SFTData[k]->fft->f0 + (REAL8)(SFTData[k]->fft->data->length - 1) * SFTData[k]->fft->deltaF);
  	fprintf(stdout, "\nfreqStepSize = %f \n", SFTData[k]->fft->deltaF);
  	fflush(stdout);
     #endif

     mddEntry->range.dFreq.startFreq = SFTData[k]->fft->f0;
     mddEntry->range.dFreq.stopFreq =  SFTData[k]->fft->f0 + (REAL8)(SFTData[k]->fft->data->length - 1) * SFTData[k]->fft->deltaF;
     mddEntry->range.dFreq.freqStepSize = SFTData[k]->fft->deltaF;

     /* mddEntry->range.dFreq.numberSamples = spec->data->length; */
     mddEntry->range.dFreq.numberSamples = SFTData[k]->fft->data->length;

     /* data type is COMPLEX8 */
     mddEntry->type = complex_8; /* 03/06/02 gam */
     /* mddEntry->type = complex_16; */ /* 09/27/01 gam */ /* 03/06/02 gam */

     /* size of the data */
     mddEntry->numberDimensions = 1;
     mddEntry->dimensions = (UINT4 *) LALMalloc (sizeof(UINT4));
     /* Below trying LALCalloc instead */
     /* mddEntry->dimensions = (UINT4 *) LALCalloc (1, sizeof(UINT4)); */
     if (!(mddEntry->dimensions))
     {
      ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
     }
     *(mddEntry->dimensions) = SFTData[k]->fft->data->length;

     /* dcHistory is ? */
     /* mddEntry->history = NULL; */ /* 11/02/01 gam */

     /* 11/02/01 gam */
     /* Set up the history structure with meta data about the SFT */

     /* dcHistory data structure */
     /*
	typedef struct dcHistoryTag
	{
   		struct dcHistoryTag* previous;
   		CHAR name[maxHistoryName];
   		CHAR units[maxHistoryUnits];
   		datatype type;
   		UINT4 numberValues;
   		dataPointer value;
   		struct dcHistoryTag* next;
	}dcHistory;
      */

      /* 11/05/01 gam */
      /* Start history linked list */
      historyEntry = mddEntry->history = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      historyEntry->previous = NULL;

      /* Use ifdef below to remove items form history list */
#ifdef HISTORY_test
      /* dso_name */
      sprintf(historyEntry->name, "dso_name" );
      sprintf(historyEntry->units, " " );

      historyEntry->type = char_s_ptr;
      historyEntry->numberValues = strlen( params->dsoName );
      historyEntry->value.chars = (CHAR *) LALMalloc( (strlen( params->dsoName ) + 1) * sizeof(CHAR) );
      if ( ! historyEntry->value.chars )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      strcpy(historyEntry->value.chars, params->dsoName);

      /* datacon_desc */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "datacon_desc" );
      sprintf(historyEntry->units, " " );

      historyEntry->type = char_s_ptr;
      historyEntry->numberValues = strlen( params->dataConDesc );
      historyEntry->value.chars = (CHAR *) LALMalloc( (strlen( params->dataConDesc ) + 1) * sizeof(CHAR) );
      if ( ! historyEntry->value.chars )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      strcpy(historyEntry->value.chars, params->dataConDesc);

      /* percent_pad */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

#endif

      /* Put percent drop out that was zero padded in time series that produced this sft */
      sprintf(historyEntry->name, "percent_pad" );
      sprintf(historyEntry->units, "percent" );
      historyEntry->type = real_4;
      historyEntry->numberValues = 1;
      historyEntry->value.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! historyEntry->value.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.real4) = 100.0*params->percentPad;  /* 11/05/01 gam */

      /* max_power */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "max_power" );
      sprintf(historyEntry->units, "power" );

      historyEntry->type = real_4;
      historyEntry->numberValues = 1;
      historyEntry->value.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! historyEntry->value.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.real4) = (REAL4)(SFTStats[k]->pwMax);

     /* freq_pwmax */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "freq_pwmax" );
      sprintf(historyEntry->units, "hz" );

      historyEntry->type = real_4;
      historyEntry->numberValues = 1;
      historyEntry->value.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! historyEntry->value.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.real4) = (REAL4)(SFTStats[k]->freqPWMax);

     /* power_mean */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "power_mean" );
      sprintf(historyEntry->units, "power" );

      historyEntry->type = real_4;
      historyEntry->numberValues = 1;
      historyEntry->value.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! historyEntry->value.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.real4) = (REAL4)(SFTStats[k]->pwMean);

     /* power_stddev */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "power_stddev" );
      sprintf(historyEntry->units, "power" );

      historyEntry->type = real_4;
      historyEntry->numberValues = 1;
      historyEntry->value.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! historyEntry->value.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.real4) = (REAL4)(SFTStats[k]->pwStdDev);

      /* startSec */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "startSec" );
      sprintf(historyEntry->units, "s" );

      historyEntry->type = int_4s;
      historyEntry->numberValues = 1;
      historyEntry->value.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! historyEntry->value.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.int4s) = params->timeIntervals[k].startSec;

      /* startNan */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "startNan" );
      sprintf(historyEntry->units, "ns" );

      historyEntry->type = int_4s;
      historyEntry->numberValues = 1;
      historyEntry->value.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! historyEntry->value.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.int4s) = params->timeIntervals[k].startNan;

      /* stopSec */
      /* 11/17/01 gam */
      /*
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "stopSec" );
      sprintf(historyEntry->units, "s" );

      historyEntry->type = int_4s;
      historyEntry->numberValues = 1;
      historyEntry->value.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! historyEntry->value.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.int4s) = params->timeIntervals[k].stopSec;
      */

      /* stopNan */
      /* 11/17/01 gam */
      /*
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "stopNan" );
      sprintf(historyEntry->units, "ns" );

      historyEntry->type = int_4s;
      historyEntry->numberValues = 1;
      historyEntry->value.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! historyEntry->value.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.int4s) = params->timeIntervals[k].stopNan;
      */

      /* sample_rate */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "sample_rate" );
      sprintf(historyEntry->units, "hz" );

      historyEntry->type = real_4;
      historyEntry->numberValues = 1;
      historyEntry->value.real4 = (REAL4 *) LALMalloc( sizeof(REAL4) );
      if ( ! historyEntry->value.real4 )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.real4) = (REAL4)(1.0/params->deltaT);

      /* ndata_input */ /* 11/17/01 gam */
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "ndata_input" );
      sprintf(historyEntry->units, " " );

      historyEntry->type = int_4s;
      historyEntry->numberValues = 1;
      historyEntry->value.int4s = (INT4 *) LALMalloc( sizeof(INT4) );
      if ( ! historyEntry->value.int4s )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      *(historyEntry->value.int4s) = params->nSFT;

      /* channel */ /* 11/18/01 gam Will try to include this later */
      /*
      historyEntry->next = (dcHistory *) LALMalloc( sizeof(dcHistory) );
      if ( ! historyEntry->next  )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      historyEntry->next->previous = historyEntry;
      historyEntry = historyEntry->next;

      sprintf(historyEntry->name, "channel" );
      sprintf(historyEntry->units, " " );

      historyEntry->type = char_s_ptr;
      historyEntry->numberValues = strlen( params->channel );
      historyEntry->value.chars = (CHAR *) LALMalloc( (strlen( params->channel ) + 1) * sizeof(CHAR) );
      if ( ! historyEntry->value.chars )
      {
        ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
      }
      strcpy(historyEntry->value.chars, params->channel);
      */

      historyEntry->next = NULL;
      /* End history linked list */

     /* allocate memory for the data */
     mddEntry->data.complex8 = (COMPLEX8 *) LALCalloc (SFTData[k]->fft->data->length, sizeof(COMPLEX8)); /* 03/06/02 gam */
     /* mddEntry->data.complex16 = (COMPLEX16 *) LALCalloc (SFTData[k]->fft->data->length, sizeof(COMPLEX16)); */ /* 09/27/01 gam */ /* 03/06/02 gam */
     /* below trying LALMalloc instead */
     /* mddEntry->data.complex8 = (COMPLEX8 *) LALMalloc (SFTData[k]->fft->data->length*sizeof(COMPLEX8)); */
     /* if (!(mddEntry->data.complex8)) */
     /* 09/27/01 gam */
     /* if (!(mddEntry->data.complex16)) */ /* 03/06/02 gam */
     if (!(mddEntry->data.complex8))
     {
      ABORT( status, KNOWNPULSARDEMODH_EALOC, KNOWNPULSARDEMODH_MSGEALOC );
     }

     /* and copy the data there */
     /* memcpy (mddEntry->data.complex8, SFTData[k]->fft->data->data, SFTData[k]->fft->data->length * sizeof(COMPLEX8) ); */
     /* memcpy (mddEntry->data.complex16, SFTData[k]->fft->data->data, SFTData[k]->fft->data->length * sizeof(COMPLEX16) ); */ /* 09/27/01 gam */ /* 03/06/02 gam */

     /* 03/06/02 gam */ /* 05/14/02 gam */
     /* for (icopy = 0; icopy < SFTData[k]->fft->data->length; icopy++) {
         mddEntry->data.complex8[icopy].re = (REAL4)SFTData[k]->fft->data->data[icopy].re;
         mddEntry->data.complex8[icopy].im = (REAL4)SFTData[k]->fft->data->data[icopy].im;
     } */ 
     memcpy (mddEntry->data.complex8, SFTData[k]->fft->data->data, SFTData[k]->fft->data->length * sizeof(COMPLEX8) );

  } /* end for (k = 0; k < output->numOutput; k++) */

   /* normal exit */
   DETATCHSTATUSPTR (status);
   RETURN (status);
}

/* 11/10/01 gam */ /* Calculate stats associated with frequency series */
/* 05/14/02 gam */
/* void CalcDemodPowerStats( const COMPLEX16Vector *fsVec, DemodPowerStats *stats, REAL8 fStart, REAL8 df ) */
void CalcDemodPowerStats( const COMPLEX8Vector *fsVec, DemodPowerStats *stats, REAL8 fStart, REAL8 df )
{

    UINT4 i = 0;         /* all purpose index */
    INT4 ipwMax = 0;     /* Index max power */
    REAL8 pw = 0.0;      /* power in one bin */
    REAL8 pwSum= 0.0;    /* Sum of power */
    REAL8 pwSum2 = 0.0;  /* Sum of power squared */

    stats->pwMax = 0.0;  /* Initialize */

    for(i=0;i<fsVec->length;i++)
    {
		/* pw = fsVec->data[i].re*fsVec->data[i].re + fsVec->data[i].im*fsVec->data[i].im; */ /* 05/14/02 gam */
		pw = (REAL8)(fsVec->data[i].re*fsVec->data[i].re + fsVec->data[i].im*fsVec->data[i].im);  /* 05/14/02 gam */

	        pwSum += pw;
	        pwSum2 += pw*pw;

		/* Update maximum power */
		if(stats->pwMax < pw)
		{
			stats->pwMax=pw;
			ipwMax=i;
		}
    }
    stats->freqPWMax = fStart + ((REAL8)ipwMax)*df; /* Freq associated with max power */
    stats->pwMean = pwSum/((REAL8)fsVec->length);   /* Mean power */
    /* Standard Deviation of the power = sqrt of the variance */
    stats->pwStdDev = pwSum2/((REAL8)fsVec->length) - stats->pwMean*stats->pwMean;
    if ( stats->pwStdDev > 0.0 ) {
    	stats->pwStdDev = sqrt( stats->pwStdDev );
    } else {
        stats->pwStdDev = -1.0; /* Should not happen except due to round off error in test data */
    }
}

/* 11/27/01 gam */  /* Calculate the mean value of a time series with cutoffs */
/* void CalcMeanWithCutoffs( const REAL4Vector *tInputVec, REAL4 cutoffMin, REAL4 cutoffMax, REAL4 *meanWithCuttoffs)
{
*/
/*    UINT4 i = 0;    */     /* all purpose index */
/*    REAL4 sum= 0.0; */   /* Sum of power */

    /* printf(stdout, "cutoffMax, cutoffMin = %g %g \n",cutoffMax,cutoffMin); */
/*
    for(i=0;i<tInputVec->length;i++)
    {
    	if ( tInputVec->data[i] <= cutoffMax && tInputVec->data[i] >= cutoffMin ) {
*/
	    /* fprintf(stdout, "i = %i, tInputVec->data[i] = %g \n",i,tInputVec->data[i]); */
/*
            sum += tInputVec->data[i];
        }
        *meanWithCuttoffs = sum/((REAL4)tInputVec->length);
    }
}
*/

/* 11/11/01 gam */ /* Find index of input data that corresponds to gpsStartTime */
void GetStartTimeIndex(const gpsTimeInterval *times, const UINT4 *gpsStartTime, UINT4 *iStartTime)
{
  UINT4 diffSec = *gpsStartTime - times->startSec;
  *iStartTime = floor((REAL8)((REAL8)diffSec - (REAL8)(1.0e-9*(REAL8)times->startNan))/((REAL8)times->timeStepSize));
}
