/***************************** <lalVerbatim file="StochasticApplySearchCV">
Author: Anderson, W.G., Romano, J.D., and Whelan, J.T.
$Id: StochasticApplySearch.c,v 1.22 2002/06/13 22:57:21 whelan Exp $
 **************************************************** </lalVerbatim> */

/******************************* <lalLaTeX file="StochasticApplySearchC">
\subsection{Module \texttt{StochasticApplySearch.c}}
\label{s:StochasticApplySearch.c}

Routine that implements the stochastic background search, calculating
the values of the spectrum of the optimally-filtered cross-correlation 
statistic, building the output for subsequent insertion into the 
metadata database and frames.

\subsubsection*{Prototypes}
\input{StochasticApplySearchCP}
\index{\texttt{LALApplySearch()}}

\subsection*{Description}

Implements the search for a stochastic background, calculating 
the values of the spectrum of the optimally-filtered cross-correlation 
statistic with $\Omega_{\rm gw}(f) = \Omega_0 = {\rm const}$.
The cross-correlation statistic spectrum values are then built for 
subsequent insertion into the metadata database and frames using the 
\texttt{StochasticBuildOutput} routine.

\subsection*{Algorithms}

Given time-series data from two interferometers (pre-processed by
the \texttt{datacondAPI}) and values of the optimal filter 
(calculated in \texttt{LALConditionData}), \texttt{LALApplySearch}
does the following:
%
\begin{enumerate}
\item 
Splits the input time-series data into \texttt{numSegments$\ge 10$}
segments, each of length \texttt{numPoints}.

\item 
Windows, zero-pads, and FFTs each segment of time-series data.

\item 
Calculates the values of the standard optimally-filtered 
cross-correlation statistic for each segment of data, using the 
values of the optimal filter previously calculated in 
\texttt{LALConditionData}.

\item
Coarse-grains the cross-correlation statistic spectrum output to 
the desired bandwidth and frequency resolution.

\item
Builds the cross-correlation statistic spectrum output 
(for each data segment) for subsequent insertion into the metadata 
database and frames.
\end{enumerate}

\subsection*{Uses}
 
\begin{itemize}
\item \texttt{LALCCreateVector}
\item \texttt{LALSZeroPadAndFFT}
\item \texttt{LALStochasticCrossCorrelationSpectrum}
\item \texttt{LALCCoarseGrainFrequencySeries}
\item \texttt{StochasticBuildOutput} 
\item \texttt{LALCDestroyVector}
\end{itemize}

\subsection*{Notes}

The analysis is run on {\em only} the search master node of the 
beowulf, 
since we are currently considering the simple case of stochastic 
signals with $\Omega_{\rm gw}(f)=\Omega_0 = {\rm const}$.
In the future, searches for stochastic backgrounds having different 
spectra (e.g., power law $\Omega_{\rm gw}(f)\propto f^\alpha$) 
should be run on multiple slaves (e.g., one node for each $\alpha$).

\vfill{\footnotesize\input{StochasticApplySearchCV}}

 *************************************************** </lalLaTeX> */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <lal/LALStdlib.h>
#include <lal/LALDatatypes.h>
#include <lal/Comm.h>
#include <lal/AVFactories.h>
#include <lal/TimeFreqFFT.h>
#include <lal/CoarseGrainFrequencySeries.h>
#include "StochasticApplySearch.h"

NRCSID (STOCHASTICAPPLYSEARCHC, "$Id: StochasticApplySearch.c,v 1.22 2002/06/13 22:57:21 whelan Exp $");

/************************ <lalVerbatim file="StochasticApplySearchCP"> */
void LALApplySearch(
      LALStatus             *status,
      LALSearchOutput       *output,
      LALSearchInput        *input,
      LALApplySearchParams  *applyParams
)
/*********************************************** </lalVerbatim> */
{

   INT4                                myRank;

   SZeroPadAndFFTParameters            szpfftParams;
   CZeroPadAndFFTParameters            czpfftParams;
   COMPLEX8FrequencySeries             fseries1;
   COMPLEX8FrequencySeries             fseries2;
   COMPLEX8FrequencySeries             ccSpectrum;
   COMPLEX8FrequencySeries             ccSpectrumOut;
   REAL4WithUnits                      ccStatReal;
   COMPLEX8WithUnits                   ccStatComplex;
   StochasticCrossCorrelationInput     ccInput;
   FrequencySamplingParams             cgParams;
   COMPLEX8                            *cPtr;
   COMPLEX8                            *cStopPtr;

   LALUnitPair                         unitPair;

   UINT4                               numFTPoints;

   StochasticSearchOutput              searchOutput;

   StochasticSearchParams*             params = NULL;
   StochasticRealDataSegment*          sDummySegment = NULL;
   StochasticComplexDataSegment*       cDummySegment = NULL;

   INITSTATUS (status, "LALApplySearch", STOCHASTICAPPLYSEARCHC);
   ATTATCHSTATUSPTR (status);


   /* 
    *
    * check output and parameters 
    * 
    */

   ASSERT (output, 
         status, 
         STOCHASTICAPPLYSEARCHH_ENULL, 
         STOCHASTICAPPLYSEARCHH_MSGENULL);

   ASSERT (!(output->result), 
         status, 
         STOCHASTICAPPLYSEARCHH_ENNUL, 
         STOCHASTICAPPLYSEARCHH_MSGENNUL);

   ASSERT (applyParams, 
         status, 
         STOCHASTICAPPLYSEARCHH_ENULL, 
         STOCHASTICAPPLYSEARCHH_MSGENULL);

   ASSERT (applyParams->mpiParams, 
         status, 
         STOCHASTICAPPLYSEARCHH_ENULL, 
         STOCHASTICAPPLYSEARCHH_MSGENULL);

   ASSERT (applyParams->searchParams, 
         status, 
         STOCHASTICAPPLYSEARCHH_ENULL, 
         STOCHASTICAPPLYSEARCHH_MSGENULL);


   /* cast the parameter structure */ 

   params = (StochasticSearchParams *) applyParams->searchParams;


   /* 
    *
    * perform the search
    *
    */

   /* 
    * 
    * NOTE: WE DO THE ANALYSIS ON **ONLY** THE SEARCH MASTER SINCE
    * WE CONSIDER THE SIMPLE CASE OF A SEARCH FOR A STOCHASTIC
    * BACKGROUND SIGNAL WITH OMEGA_GW=CONST.  
    * SEARCHES FOR STOCHASTIC SIGNALS HAVING DIFFERENT SPECTRA CAN 
    * BE DONE ON MULTIPLE SEARCH SLAVES.
    * THIS IS WORK FOR THE FUTURE!!
    * 
    */   

   /* if not search master, abort */
   if (!(params->searchMaster))
   {
      ABORT (status, 
            STOCHASTICAPPLYSEARCHH_EMSTR, 
            STOCHASTICAPPLYSEARCHH_MSGEMSTR);
   }

   /* determine rank and check that search master is running on node 0 */
   /* :TODO: is this correct?? */
   MPI_Comm_rank (*(applyParams->mpiParams->comm), &myRank);

   if (myRank != 0)
   {
      ABORT (status, 
            STOCHASTICAPPLYSEARCHH_ERANK, 
            STOCHASTICAPPLYSEARCHH_MSGERANK);
   }

   if (params->currentSegment < params->initParams->numSegments)
   {

     /* we still have data to analyze, so carry on with search */

     /* Assign searchOutput pointers depending on outputs requested */

     if ( ( STOCHASTICH_CCSPECDBBIT | STOCHASTICH_CCSPECFRBIT )
	  & params->initParams->outputType ) 
     {
       searchOutput.ccSpec = &ccSpectrumOut;
     }
     else
     {
       searchOutput.ccSpec = NULL;
     }

     if ( STOCHASTICH_CCSTATDBBIT & params->initParams->outputType ) 
     {
       if (!(params->initParams->baseFreq))
       { /* real option */
	 searchOutput.ccStatReal = &ccStatReal;
	 searchOutput.ccStatComplex = NULL;
       }
       else
       {
	 searchOutput.ccStatReal = NULL;
	 searchOutput.ccStatComplex = &ccStatComplex; 
       }
     }
     else
     {
       searchOutput.ccStatReal = NULL;
       searchOutput.ccStatComplex = NULL;
     }

     if (params->optimalFilter)
     {
       
       /* 
	*
	* allocate temporary memory for the frequency data 
	*
	*/
       
       numFTPoints = ( params->initParams->baseFreq 
		       ? ( params->initParams->fullLength )
		       : params->initParams->fullLength / 2 + 1
		       );
       
       fseries1.data = NULL;
       LALCCreateVector (status->statusPtr, 
			 &(fseries1.data),
			 numFTPoints);
       CHECKSTATUSPTR (status);
       
       fseries2.data = NULL;
       LALCCreateVector (status->statusPtr, 
			 &(fseries2.data), 
			 numFTPoints);
       CHECKSTATUSPTR (status);
       
       if (searchOutput.ccSpec)
       {
	 ccSpectrum.data = NULL;
	 LALCCreateVector (status->statusPtr, 
			   &(ccSpectrum.data), 
			   params->initParams->numFrequencies);
	 CHECKSTATUSPTR (status);
	 
	 ccSpectrumOut.data = NULL;
	 LALCCreateVector (status->statusPtr, 
			   &(ccSpectrumOut.data), 
			   params->initParams->outputNumFreq);
	 CHECKSTATUSPTR (status);
       }
       
       /* 
	* zero-pad and FFT the time-series data
	*/
       
       if (!(params->initParams->baseFreq))
       { /* real option */
	 /* point dummySegment to the current segment to analyze */
	 sDummySegment 
	   = params->stochasticSegVec.realVec->data + params->currentSegment;
	 
	 szpfftParams.fftPlan = params->fftPlan.realPlan;
	 szpfftParams.length  = params->initParams->fullLength;
	 szpfftParams.window  = params->window1;
	 LALSZeroPadAndFFT (status->statusPtr,
			    &fseries1,
			    sDummySegment->data1,
			    &szpfftParams);
	 CHECKSTATUSPTR (status);
	 
	 szpfftParams.window = params->window2;
	 LALSZeroPadAndFFT (status->statusPtr,
			    &fseries2,
			    sDummySegment->data2,
			    &szpfftParams);
	 CHECKSTATUSPTR (status);
	 
       } /* end real option */
       else
       { /* complex option */
	 /* point dummySegment to the current segment to analyze */
	 cDummySegment 
	   = params->stochasticSegVec.complexVec->data + params->currentSegment;
	 
	 czpfftParams.fftPlan = params->fftPlan.complexPlan;
	 czpfftParams.length  = params->initParams->fullLength;
	 czpfftParams.window  = params->window1;
	 LALCZeroPadAndFFT (status->statusPtr,
			    &fseries1,
			    cDummySegment->data1,
			    &czpfftParams);
	 CHECKSTATUSPTR (status);
	 
	 czpfftParams.window  = params->window2;
	 LALCZeroPadAndFFT (status->statusPtr,
			    &fseries2,
			    cDummySegment->data2,
			    &czpfftParams);
	 CHECKSTATUSPTR (status);
       }

       /* 
	*
	* build stochastic cross-correlation input structure 
	*
	*/
       
       ccInput.hBarTildeOne  = &fseries1;
       ccInput.hBarTildeTwo  = &fseries2;
       ccInput.optimalFilter = params->optimalFilter;
       
       if ( searchOutput.ccSpec ) 
       {
	 /* 
	  * calculate the cross-correlation spectrum 
	  */
	 LALStochasticCrossCorrelationSpectrum (status->statusPtr,
						&ccSpectrum,
						&ccInput,
						params->initParams->epochsMatch);
	 CHECKSTATUSPTR (status);
 
	 cgParams.length = params->initParams->outputNumFreq;
	 cgParams.f0     = params->initParams->outputF0;
	 cgParams.deltaF = params->initParams->outputDeltaF;
	 
	 /* Coarse-grain the output to the desired resolution */
	 TRY( LALCCoarseGrainFrequencySeries(status->statusPtr,
					     &ccSpectrumOut,
					     &ccSpectrum,
					     &cgParams),
	      status);
	 
	 /* 
	  *
	  * build output for metadata database or trend frames 
	  *
	  */
	 
	 if ( searchOutput.ccStatReal )
	 { /* real option */
	   if (ccSpectrum.f0 == 0) 
	   {
	     /* DC contribution */
	     ccStatReal.value 
	       = ccSpectrum.data->data->re / 2.0;
	     
	     /* We might want to check that the imaginary parts of the DC
		components of all the series vanish */
	     
	     /* initialize pointers */
	     cPtr = ccSpectrum.data->data + 1;
	   } /* if f0 == 0 */
	   else 
	   {
	     ccStatReal.value  = 0.0;
	     
	     /* initialize pointers */
	     cPtr = ccSpectrum.data->data;
	   }    
	   cStopPtr    = ccSpectrum.data->data + ccSpectrum.data->length;
	   /* contributions from positive and (negative) frequency components */
	   for ( ; cPtr < cStopPtr; ++cPtr ) 
	   {
	     ccStatReal.value  += cPtr->re;
	   }
	   
	   /* normalize */
	   ccStatReal.value  *= ccSpectrum.deltaF * 2.0 ;
	   
	   /* Set output units */  
	   unitPair.unitOne = &ccSpectrum.sampleUnits;
	   unitPair.unitTwo =  &lalHertzUnit;
	   TRY( LALUnitMultiply(status->statusPtr,
				&(ccStatReal.units), 
				&unitPair),
		status );
	 } /* end real option */
	 else if ( searchOutput.ccStatComplex )
	 { /* complex option */
	   
	   ccStatComplex.value.re = ccStatComplex.value.im = 0.0;
	   
	   /* initialize pointers */
	   cPtr = ccSpectrum.data->data;
	   
	   cStopPtr    = ccSpectrum.data->data + ccSpectrum.data->length;
	   /* contributions from positive and (negative) frequency components */
	   for ( ; cPtr < cStopPtr; ++cPtr ) 
	   {
	     ccStatComplex.value.re += cPtr->re;
	     ccStatComplex.value.im += cPtr->im;
	   }
	   
	   /* normalize */
	   ccStatComplex.value.re *= ccSpectrum.deltaF;
	   ccStatComplex.value.im *= ccSpectrum.deltaF;
	   
	   /* Set output units */  
	   unitPair.unitOne = &ccSpectrum.sampleUnits;
	   unitPair.unitTwo =  &lalHertzUnit;
	   TRY( LALUnitMultiply(status->statusPtr, &(ccStatComplex.units), &unitPair),
		status );
	 }

	 /* 
	  * remove temporary memory for the frequency data 
	  */
	 
	 LALCDestroyVector (status->statusPtr, &(ccSpectrum.data));
	 CHECKSTATUSPTR (status);

       }
       else if ( searchOutput.ccStatReal )
       { /* real option */
	 LALStochasticCrossCorrelationStatistic
	   (status->statusPtr,
	    &ccStatReal,
	    &ccInput,
	    params->initParams->epochsMatch);
       } /* end real option */
       else if  ( searchOutput.ccStatComplex )
       { /* complex option */
	 LALStochasticHeterodynedCrossCorrelationStatistic
	   (status->statusPtr,
	    &ccStatComplex,
	    &ccInput,
	    params->initParams->epochsMatch);
       }
     
       /* 
       	* remove temporary memory for the frequency data 
	*/
       
       LALCDestroyVector (status->statusPtr, &(fseries2.data));
       CHECKSTATUSPTR (status);
       
       LALCDestroyVector (status->statusPtr, &(fseries1.data));
       CHECKSTATUSPTR (status);
     }

     if ( params->currentSegment == 0 )
     {
       searchOutput.sigma = params->sigma;
     }
     else
     {
       searchOutput.sigma = NULL;
     }

     searchOutput.outputType = params->initParams->outputType;

     StochasticBuildOutput(status->statusPtr,
			   output,
			   &searchOutput,
			   params);
     CHECKSTATUSPTR (status);

     if ( searchOutput.ccSpec )
     {
       TRY( LALCDestroyVector (status->statusPtr, &(ccSpectrumOut.data)),
	    status );
     }

     /*
      *  
      * compute fraction of search remaining for wrapper and set 
      * output not finished 
      * 
      * :NOTE: this is what duncan did for simple.so even though
      * it would seem that i should increment to the pointer
      * to the current segment before computing fracRemaining and
      * before determining how to set notFinished.
      *
      */  
     
     output->fracRemaining = 1.0 - 
       (REAL4)(params->currentSegment)/ 
       (REAL4)(params->initParams->numSegments);
     
     output->notFinished = STOCHASTICH_TRUE;
     
     
     /* increment pointer to current segment */
     (params->currentSegment)++;
     
   }
   else
   {
      /* we are finished with the search */

      output->fracRemaining = 0;
      output->notFinished   = STOCHASTICH_FALSE;
   }


   /* normal exit */ 
   DETATCHSTATUSPTR (status);
   RETURN (status);
}

